# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
r"""
Defines a class with common functions for MaNGA fits files.  This class
needs to be as core as possible (little to no dependencies are
higher-level classes).

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import sys
import os
import numpy
import logging
import warnings

from IPython import embed

from scipy import sparse

from astropy.wcs import WCS
from astropy.io import fits
import astropy.constants

from mangadap import __version__

from .bitmask import BitMask
from .log import log_output
from .fileio import compress_file, create_symlink
from .pixelmask import SpectralPixelMask

class DAPFitsUtil:
    r"""
    An abstract class that only implements a set of static methods for
    interaction with MaNGA fits files.

    .. todo::
        - Make this a base class for all the fits files used/created by
          the DAP?
        - Use HDUList as a base class for this object?

    """
    @staticmethod
    def storage_mode_options():
        """
        Return the allowed storage modes.

        Returns:
            list: List of the allowed storage mode for DAP fits files.
        """
        return [ '2d', '3d' ]


    @staticmethod
    def get_spatial_shape(shape, dispaxis):
        """
        
        Return the shape of the spatial axes given the specified
        dispersion axis.

        Args:
            shape (tuple): Shape of the full numpy.ndarray object.
            dispaxis (int): Axis with the spectra.

        Returns:
            tuple : Shape of the remaining axes.
        """
        return shape[:dispaxis] + shape[dispaxis+1:]


    @staticmethod
    def clean_dap_primary_header(hdr):
        # Remove some keys that are incorrect for DAP data
        hdr.remove('BSCALE', ignore_missing=True)
        hdr.remove('BZERO', ignore_missing=True)
        hdr.remove('BUNIT', ignore_missing=True)
        hdr.remove('MASKNAME', ignore_missing=True)
        return hdr


    @staticmethod
    def initialize_dap_primary_header(cube, maskname=None):
        # Copy the from the DRP and clean it
        hdr = cube.prihdr.copy()
        hdr = DAPFitsUtil.clean_dap_primary_header(hdr)

        # Change MASKNAME
        if maskname is not None:
            hdr['MASKNAME'] = maskname

        # Add versioning
        hdr['VERSPY'] = ('.'.join([ str(v) for v in sys.version_info[:3]]), 'Python version')
        hdr['VERSNP'] = (numpy.__version__, 'Numpy version')
        import scipy
        hdr['VERSSCI'] = (scipy.__version__, 'Scipy version')
        import astropy
        hdr['VERSAST'] = (astropy.__version__, 'Astropy version')
        import pydl
        hdr['VERSPYDL'] = (pydl.__version__, 'pydl version')
        import ppxf
        hdr['VERSPPXF'] = (ppxf.__version__, 'pPXF version')
        try:
            import vorbin
            hdr['VERSVBIN'] = (vorbin.__version__, 'VorBin version (if used)')
        except:
            pass
        hdr['VERSDAP'] = (__version__, 'MaNGA DAP version')

        return hdr


    @staticmethod
    def finalize_dap_header(hdr, ext, bunit=None, hduclas2='DATA', err=False, qual=False,
                            multichannel=False, bit_type=None, prepend=True, channel_names=None,
                            channel_units=None):

        _hdr = hdr.copy() if channel_names is None else \
                    DAPFitsUtil.add_channel_names(hdr, channel_names, units=channel_units)

        if bunit is not None:
            _hdr['BUNIT'] = (bunit, 'Unit of pixel value')

        # Add the common HDUCLASS keys
        _hdr['HDUCLASS'] = ('SDSS', 'SDSS format class')
        _hdr['HDUCLAS1'] = ('CUBE' if multichannel else 'IMAGE', 'Data format')
        if hduclas2 == 'DATA':
            _hdr['HDUCLAS2'] = 'DATA'
            if err:
                _hdr['ERRDATA'] = (ext+'_IVAR' if prepend else 'IVAR',
                                    'Associated inv. variance extension')
            if qual:
                _hdr['QUALDATA'] = (ext+'_MASK' if prepend else 'MASK',
                                    'Associated quality extension')
            return _hdr

        if hduclas2 == 'ERROR':
            _hdr['HDUCLAS2'] = 'ERROR'
            _hdr['HDUCLAS3'] = ('INVMSE', 'Value is inverse mean-square error')
            _hdr['SCIDATA'] = (ext, 'Associated data extension')
            if qual:
                _hdr['QUALDATA'] = (ext+'_MASK' if prepend else 'MASK',
                                    'Associated quality extension')
            return _hdr

        if hduclas2 == 'QUALITY':
            _hdr['HDUCLAS2'] = 'QUALITY'
            _hdr['HDUCLAS3'] = DAPFitsUtil._mask_data_type(bit_type)
            _hdr['SCIDATA'] = (ext, 'Associated data extension')
            if err:
                _hdr['ERRDATA'] = (ext+'_IVAR' if prepend else 'IVAR',
                                    'Associated inv. variance extension')
            return _hdr
            
        raise ValueError('HDUCLAS2 must be DATA, ERROR, or QUALITY.')


    @staticmethod
    def clean_map_header(hdr, multichannel=False):

        # Change header keywords to the default values for the third axis
        # TODO: Should instead read the header into a wcs, manipulate the wcs
        # object and then rewrite it to a header.
        if multichannel:
            hdr['NAXIS'] = 3
            if 'CTYPE3' in hdr:
                hdr.remove('CTYPE3')
            if 'CUNIT3' in hdr:
                hdr.remove('CUNIT3')
            hdr['CTYPE3'] = ' '
            hdr['CUNIT3'] = ' '
            hdr['CRPIX3'] = 1
            hdr['CRVAL3'] = 1.
            hdr['CD3_3']  = 1.
        else:
            hdr['NAXIS'] = 2
            #hdr.remove('NAXIS3')
            for key in ['CTYPE3', 'CUNIT3', 'CRPIX3', 'CRVAL3', 'CD3_3', 'CDELT3']:
                if key in hdr:
                    hdr.remove(key)

        # Remove everything but the WCS information
        w = WCS(header=hdr)
        hdr = w.to_header().copy()

        # Fix the DATE-OBS keyword:
        if 'DATA-OBS' in hdr:
            hdr.comments['DATE-OBS'] = 'Date of median exposure'
        if 'MJD-OBS' in hdr:
            hdr.comments['MJD-OBS'] = '[d] MJD for DATE-OBS'

        # Add back in the BSCALE and BZERO values
        hdr['BSCALE'] = 1.
        hdr['BZERO'] = 0.

        # Add back in the default CTYPE3, CUNIT3
        if multichannel:
            hdr['CTYPE3'] = (' ', 'Undefined type')
            hdr['CUNIT3'] = (' ', 'Undefined units')

        return hdr


    @staticmethod
    def build_map_header(hdr, author, multichannel=False, maskname=None):
        hdr = hdr.copy()
        hdr = DAPFitsUtil.clean_map_header(hdr, multichannel=multichannel)
        hdr['AUTHOR'] = author
        if maskname is not None:
            hdr['MASKNAME'] = maskname
        return hdr


    @staticmethod
    def add_channel_names(hdr, names, units=None):
        _hdr = hdr.copy()
        nchannels = len(names)
        ndig = int(numpy.log10(nchannels))+1
        if units is not None and len(units) != nchannels:
            raise ValueError('Length of column names and units are not the same.')

        for i in range(nchannels):
            _hdr['C'+'{0}'.format(i+1).zfill(ndig)] = (names[i], 'Data in channel {0}'.format(i+1))
            if units is not None:
                _hdr['U'+'{0}'.format(i+1).zfill(ndig)] \
                            = (units[i], 'Units of data in channel {0}'.format(i+1))
        return _hdr


    @staticmethod
    def clean_cube_header(hdr):

        # Remove everything but the WCS information
        w = WCS(header=hdr)
        hdr = w.to_header().copy()

        # Fix the DATE-OBS keyword:
        if 'DATA-OBS' in hdr:
            hdr.comments['DATE-OBS'] = 'Date of median exposure'
        if 'MJD-OBS' in hdr:
            hdr.comments['MJD-OBS'] = '[d] MJD for DATE-OBS'

        # Add back in the BSCALE and BZERO values; BUNIT added during
        # "finalize"
        hdr['BSCALE'] = 1.
        hdr['BZERO'] = 0.

        return hdr


#    @staticmethod
#    def build_cube_header(drpf, author, maskname=None):
#        hdr = drpf.hdu['FLUX'].header.copy()
#        hdr = DAPFitsUtil.clean_cube_header(hdr)
#        hdr['AUTHOR'] = author
#        if maskname is not None:
#            hdr['MASKNAME'] = maskname
#        return hdr


    @staticmethod
    def build_cube_header(cube, author, maskname=None):
        hdr = cube.fluxhdr.copy()
        hdr = DAPFitsUtil.clean_cube_header(hdr)
        hdr['AUTHOR'] = author
        if maskname is not None:
            hdr['MASKNAME'] = maskname
        return hdr


    @staticmethod
    def downselect_bins(bin_indx, good_bins):
        """
        Replace values in bin_indx with -1 if they are not in good_bins.
        """
        unique_bins, reconstruct = numpy.unique(bin_indx, return_inverse=True)
        missing_bins = list(set(unique_bins) - set(good_bins))
        unique_bins = numpy.array([ -1 if x in missing_bins else x for x in unique_bins])
        return unique_bins[reconstruct]


    @staticmethod
    def match_image_naxis(hdu, ext=None):
        """
        Match the NAXISn header keywords to the shape of the image data.
        That is, reset the header keywrods such that the shape of the
        data array is (NAXIS1, NAXIS2, NAXIS3, ...). 

        Note that any reordering of the NAXISn keywords will *not*
        change the WCS information.  You will have to account for if you
        need the WCS information.

        .. todo::
            If a WCS is detected, use WCS.swapaxes() to transpose the
            WCS system!
        """
        _ext = [ h.name for h in hdu ] if ext is None else ext
        for e in _ext:
            if isinstance(hdu[e], (fits.hdu.image.PrimaryHDU, fits.hdu.image.ImageHDU)) \
                    and hdu[e].data is not None:
                shape = hdu[e].data.shape
                for i in range(len(shape)):
                    hdu[e].header['NAXIS{0}'.format(i+1)] = shape[i]
        return hdu


    @staticmethod
    def transpose_image_data(hdu, ext=None):
        """
        Transpose all image data in an HDUList.

        The primary use of this function is to ensure the HDUList data
        arrays use C-contiguous memory but with the natural array
        ordering of the axes that follow the NAXIS keywords in the
        header.  I.e., data arrays provided by astropy.io.fits have
        shapes like (..., NAXIS3, NAXIS2, NAXIS1).  This function
        reorders the data arrays to have shapes like (NAXIS1, NAXIS2,
        NAXIS3, ...), and ensures that the NAXISn header keywords
        reflect this.  Typically, one would do::

            hdu = transpose_image_data(fits.open('test.fits'))

        To prepare the image data for writing to a new fits file, one
        would do::
        
            transpose_image_data(hdu).writeto('new.fits')

        If you just want to ensure that the NAXISn keywords refect the
        python native ordering, then use :func:`match_image_naxis`.

        Note that any WCS information will **not** be transposed, which
        you will have to account for if you need the WCS information.

        .. todo::
            If a WCS is detected, use WCS.swapaxes() to transpose the
            WCS system!

        """
        _ext = [ h.name for h in hdu ] if ext is None else ext
        for e in _ext:
            if isinstance(hdu[e], (fits.hdu.image.PrimaryHDU, fits.hdu.image.ImageHDU)) \
                    and hdu[e].data is not None:
                hdu[e].data = numpy.asarray(hdu[e].data.T, order='C')
                shape = hdu[e].data.shape
                for i in range(len(shape)):
                    hdu[e].header['NAXIS{0}'.format(i+1)] = shape[i]
        return hdu


    @staticmethod
    def reconstruct_map(map_shape, bin_indx, arr, dtype=None, quiet=False):
        r"""
        Reconstruct a set of maps with the specified shape based on a
        set of input 1D arrays.  The map shape is expected to be
        :math:`(N_x,N_y)` and the array shapes are expected to be
        :math:`(N_{\rm pix},)`, where :math:`N_{\rm pix}` does *not*
        have to match the number of unique bins.  The bin_indx array is
        expected to be a vector with length :math:`N_x\timesN_y`.

        The data type of the output arrays is set by the input type if
        the provided dtype is None; otherwise, the code attempts to case
        each output map according to each element of dtype.

        Returns a tuple if the number of arrays is larger than one so
        that they can be caught by individual variables.
        """
        # Check the input types
        if not isinstance(map_shape, tuple):
            raise TypeError('Map shape must be a tuple.')
        if not isinstance(bin_indx, numpy.ndarray):
            raise TypeError('Input bin_indx must be a numpy.ndarray.')
        if not isinstance(arr, (list, tuple, numpy.ndarray)):
            raise TypeError('Input arr must be a list or tuple of numpy.ndarray objects or a '
                            'single numpy.ndarray.')
        _arr = [arr] if isinstance(arr, numpy.ndarray) else list(arr)
        if dtype is None:
            _dtype = []
            for a in _arr:
                _dtype += [a.dtype.name]
        else:
            _dtype = [dtype] if isinstance(dtype, str) else list(dtype)

        # Check the input sizes
        if len(bin_indx.shape) != 1:
            raise ValueError('Input bin_indx must be a 1D vector.')
        if bin_indx.size != numpy.prod(map_shape):
            raise ValueError('Input bin_indx length does not match cube shape.')
        narr = len(_arr)
        if len(_dtype) != narr:
            raise ValueError('Number of dtypes must match the number of arrays.')
        for i in range(1,narr):
            if _arr[i].shape != _arr[i-1].shape:
                raise ValueError('All arrays must have the same shape.')

        # Get the unique bins and how to reconstruct the bins from the
        # unique set
        unique_bins, reconstruct = numpy.unique(bin_indx, return_inverse=True)
#        print(unique_bins)

        # Handle missing bins by changing from the bin id to the bin index.
        # Detecting whether or not there is a need for handling missing bins is
        # done by comparing the unique array to a full string of all indices up
        # to the maximum index in the unique array.  **This assumes bins can
        # only either be -1, to indicate that the spaxel/bin was not analyzed,
        # or a non-negative index number** (i.e., all bin IDs must be >= -1).
        # The code handles cases both with and without any ignored bins.
        warn = False
        if numpy.any(unique_bins < 0):
            # Includes ignored bins/spaxels
            if unique_bins.size != unique_bins[-1]+2 \
                    or numpy.any((unique_bins - numpy.arange(-1,unique_bins.size-1)) != 0):
                warn = True
                unique_bins = numpy.arange(-1,unique_bins.size-1)
        else:
            # All bins/spaxels have valid bin numbers
            if unique_bins.size != unique_bins[-1]+1 \
                    or numpy.any((unique_bins - numpy.arange(unique_bins.size)) != 0):
                warn = True
                unique_bins = numpy.arange(unique_bins.size)
        if warn and not quiet:
            warnings.warn('Bin numbers and indices do not match.  Map values are expected to be '
                          'sorted by their bin number.')

#        if unique_bins.size != unique_bins[-1]+2 \
#                or numpy.any((unique_bins - numpy.arange(-1,unique_bins.size-1)) != 0):
#            if not quiet:
#                warnings.warn('Bin numbers and indices do not match.  Map values are expected '
#                              'to be sorted by their bin number.')
#            unique_bins = numpy.arange(-1,unique_bins.size-1)

        # Get the valid bins
        indx = bin_indx > -1
#        print(map_shape, bin_indx.shape, numpy.sum(indx), numpy.prod(bin_indx.shape))

        # Restructure all the arrays to match the cube shape
        new_arr = numpy.empty(narr, dtype=object)
#        print('narr:', narr)
        for i in range(narr):
            new_arr[i] = numpy.zeros(map_shape, dtype=_dtype[i])
            new_arr[i].ravel()[indx] = _arr[i][unique_bins[reconstruct[indx]]].copy()
            s = numpy.sum(numpy.invert(numpy.isfinite(new_arr[i])))
            if s > 0:
                print(i+1, narr)
                print('dtype: ', _dtype[i])
                a = numpy.sum(numpy.invert(numpy.isfinite(_arr[i][unique_bins[reconstruct[indx]]])))
                print('not finite input:', a)
                print('not finite output ravel:',
                      numpy.sum(numpy.invert(numpy.isfinite(new_arr[i].ravel()[indx]))))
                print('not finite output:', s)
                nf = numpy.invert(numpy.isfinite(new_arr[i].ravel()[indx]))
                print('inp: ', _arr[i][unique_bins[reconstruct[indx]]][nf])
                print('out: ', new_arr[i].ravel()[indx][nf])
                new_arr[i].ravel()[indx][nf] = 0.0
                raise ValueError('NaNs in data!')

        return tuple([ a for a in new_arr]) if narr > 1 else new_arr[0]


    @staticmethod
    def reconstruct_cube(cube_shape, bin_indx, arr, dtype=None):
        r"""
        Reconstruct a set of cubes with the specified shape based on a
        set of input 2D arrays.  The cube shape is expected to be
        :math:`(N_x,N_y,N_\lamba)` and the array shapes are expected to
        be :math:`(N_{\rm spec},N_\lambda)`, where :math:`N_{\rm spec}`
        does *not* have to match the number of unique bins.  The
        bin_indx array is expected to be a vector with length
        :math:`N_x\timesN_y`.

        The data type of the output arrays is set by the input type if
        the provided dtype is None; otherwise, the code attempts to case
        each output cube accordint to each element of dtype.

        Returns a tuple if the number of arrays is larger than one so
        that they can be caught by individual variables.

        """

        # Check the input types
        if not isinstance(cube_shape, tuple):
            raise TypeError('Cube shape must be a tuple.')
        if not isinstance(bin_indx, numpy.ndarray):
            raise TypeError('Input bin_indx must be a numpy.ndarray.')
        if not isinstance(arr, (list, tuple, numpy.ndarray)):
            raise TypeError('Input arr must be a list or tuple of numpy.ndarray objects or a '
                            'single numpy.ndarray.')
        _arr = [arr] if isinstance(arr, numpy.ndarray) else list(arr)
        if dtype is None:
            _dtype = []
            for a in _arr:
                _dtype += [a.dtype.name]
        else:
            _dtype = [dtype] if isinstance(dtype, str) else list(dtype)

        # Check the input sizes
        if len(bin_indx.shape) != 1:
            raise ValueError('Input bin_indx must be a 1D vector.')
        if bin_indx.size != numpy.prod(cube_shape[0:2]):
            raise ValueError('Input bin_indx length does not match cube shape.')
        narr = len(_arr)
        if len(_dtype) != narr:
            raise ValueError('Number of dtypes must match the number of arrays.')
        nwave = _arr[0].shape[-1]
        if cube_shape[-1] != nwave:
            raise ValueError('Number of pixels in arrays must match last dimension of cube.')
        for i in range(1,narr):
            if _arr[i].shape != _arr[i-1].shape:
                raise ValueError('All arrays must have the same shape.')

        # Get the unique bins and how to reconstruct the bins from the
        # unique set
        unique_bins, reconstruct = numpy.unique(bin_indx, return_inverse=True)

        # Handle missing bins by changing from the bin id to the bin index.
        # Detecting whether or not there is a need for handling missing bins is
        # done by comparing the unique array to a full string of all indices up
        # to the maximum index in the unique array.  **This assumes bins can
        # only either be -1, to indicate that the spaxel/bin was not analyzed,
        # or a non-negative index number** (i.e., all bin IDs must be >= -1).
        # The code handles cases both with and without any ignored bins.
        warn = False
        if numpy.any(unique_bins < 0):
            # Includes ignored bins/spaxels
            if unique_bins.size != unique_bins[-1]+2 \
                    or numpy.any((unique_bins - numpy.arange(-1,unique_bins.size-1)) != 0):
                warn = True
                unique_bins = numpy.arange(-1,unique_bins.size-1)
        else:
            # All bins/spaxels have valid bin numbers
            if unique_bins.size != unique_bins[-1]+1 \
                    or numpy.any((unique_bins - numpy.arange(unique_bins.size)) != 0):
                warn = True
                unique_bins = numpy.arange(unique_bins.size)
        if warn and not quiet:
            warnings.warn('Bin numbers and indices do not match.  Map values are expected to be '
                          'sorted by their bin number.')

#        if unique_bins.size != unique_bins[-1]+2 \
#                or numpy.any((unique_bins - numpy.arange(-1,unique_bins.size-1)) != 0):
#            warnings.warn('Bin numbers and indices do not match.  Spectra are expected '
#                          'to be sorted by their bin number.')
#            unique_bins = numpy.arange(-1,unique_bins.size-1)

        # Get the valid bins
        indx = bin_indx > -1

        # Restructure all the arrays to match the cube shape
        new_arr = numpy.empty(narr, dtype=object)
        for i in range(narr):
            new_arr[i] = numpy.zeros(cube_shape, dtype=_dtype[i])
            new_arr[i].reshape(-1,nwave)[indx,:] = _arr[i][unique_bins[reconstruct[indx]],:]

        return tuple([ a for a in new_arr]) if narr > 1 else new_arr[0]


    @staticmethod
    def read(ofile, permissions='readonly', checksum=False):
        return DAPFitsUtil.transpose_image_data(fits.open(ofile, mode=permissions,
                                                          checksum=checksum))


    @staticmethod
    def write(hdu, ofile, overwrite=False, checksum=False, symlink_dir=None, relative_symlink=True,
              loggers=None, quiet=False):
        """
        Write an HDUList to an output file.  It is expected that the hdu
        was either read by the :func:`read` function, or that the hdu
        was passed through :func:`transpose_image_data` such that
        HDUList data arrays use C-contiguous memory but with the natural
        array ordering of the axes that follow the NAXIS keywords in the
        header.
        """
        # Get the output file and determine if it should be compressed
        compress = False
        if ofile.split('.')[-1] == 'gz':
            _ofile = ofile[:ofile.rfind('.')] 
            compress = True
        else:
            _ofile = ofile
    
        # Transpose the order
        hdu = DAPFitsUtil.transpose_image_data(hdu)

        # Write the data
        if not quiet:
            log_output(loggers, 1, logging.INFO, 'Writing: {0}'.format(_ofile))
        hdu.writeto(_ofile, overwrite=overwrite, checksum=checksum)

        # Transpose it back
        hdu = DAPFitsUtil.transpose_image_data(hdu)

        # Compress if desired
        if compress:
            if not quiet:
                log_output(loggers, 1, logging.INFO, 'Compressing: {0}'.format(ofile))
            # And compress it
            compress_file(_ofile, overwrite=overwrite)
            os.remove(_ofile)
    
        # Create the symlink if requested
        if symlink_dir is not None:
            create_symlink(ofile, symlink_dir, relative_symlink=relative_symlink, loggers=loggers,
                           quiet=quiet)


    @staticmethod
    def unique_bins(bin_indx, return_index=False):
        """
        Get the unique bins and the indices of the unique bins in the
        flattened spatial dimension, ignoring the bins with indices of
        -1.
        """
        uniq, indx = numpy.unique(bin_indx.ravel(), return_index=True)
        if uniq[0] == -1:
            uniq = uniq[1:]
            indx = indx[1:]
        return (uniq, indx) if return_index else uniq
        
#        return tuple(map(lambda x : x[1:], numpy.unique(bin_indx.ravel(), return_index=True))) \
#                    if return_index else numpy.unique(bin_indx.ravel())[1:]


    # TODO: Consolidate these with what's in datacube and rowstackedspectra.
    @staticmethod
    def copy_to_array(hdu, ext='FLUX', allowed_ext=None, waverange=None, select_bins=None,
                      missing_bins=None, nbins=None, unique_bins=None):
        r"""
        Return a copy of the selected data array.  The array size is
        always :math:`N_{\rm spec} \times N_{\rm wavelength}`; i.e., the
        CUBE files are flattened to two dimensions, matching the
        dimensionality of the RSS files.  The spatial positions within
        the original DRP file for each spectrum are given by tuples in
        :attr:`spatial_index`.  See :func:`copy_to_masked_array` for
        arguments.

        Returns:
            numpy.ndarray: A 2D array with a copy of the data from the
            selected extension.
        """
        masked_data = DAPFitsUtil.copy_to_masked_array(hdu, ext=ext, allowed_ext=allowed_ext,
                                                       waverange=waverange, select_bins=select_bins,
                                                       missing_bins=missing_bins, nbins=nbins,
                                                       unique_bins=unique_bins)
        # For this approach, the wavelengths masked should be
        # *identical* for all spectra
        nwave = numpy.sum(numpy.invert(numpy.ma.getmaskarray(masked_data))[0,:])
        if nwave == 0:
            raise ValueError('Full wavelength range has been masked!')
        # Compression returns a flattened array, so it needs to be
        # reshaped for the new number of (unmasked) wavelength channels
        return masked_data.compressed().reshape(-1,nwave)


    @staticmethod
    def copy_to_masked_array(hdu, ext='FLUX', mask_ext=None, flag=None, bitmask=None,
                             allowed_ext=None, waverange=None, select_bins=None, missing_bins=None,
                             nbins=None, unique_bins=None):
        r"""
        Return a copy of the selected data array as a masked array.
        This is functionally identical to :func:`DRPFits.copy_to_array`,
        except the output format is a `numpy.ma.MaskedArray`_.  The
        pixels that are considered to be masked can be specified using
        the `flag` option.
        
        Args:
            waverange (array-like) : (**Optional**) Two-element array
                with the first and last wavelength to include in the
                computation.  Default is to use the full wavelength
                range.
            ext (str) : (**Optional**) Name of the extension from which
                to draw the data.  Must be allowed for the current
                :attr:`mode`; see :attr:`spectral_arrays`.  Default is
                `'FLUX'`.
            mask_ext (str) : (**Optional**) Name of the extension with
                the mask bit data.  Must be allowed for the current
                :attr:`mode`; see :attr:`spectral_arrays`.  Default is
                `'MASK'`.
            flag (str or list): (**Optional**) (List of) Flag names that
                are considered when deciding if a pixel should be
                masked.  The names *must* be a valid bit name as defined
                by the provided bitmask (see
                :class:`mangadap.util.bitmask.BitMask`).  If not
                provided, *ANY* non-zero mask bit is omitted.
            bitmask (:class:`mangadap.util.bitmask.BitMask`):
                (**Optional**) BitMask object used to determine if pixel
                is flagged as being masked by a specific maskbit.
            allowed_ext (array-like): (**Optional**) List of allowed
                extenstions that can be copied to an array; the mask
                extension **must** be one of these.
            waverange (array-like): (**Optional**) A two-element
                array with the valid wavelength range.  Anything outside
                of this range is masked.
            select_bins (array-like): (**Optional**) A list of spectra
                in the flattened cube (with shape :math:`N_{\rm spec}
                \times N_\lambda`) to return.
            missing_bins (list): (**Optional**) A list of bin numbers
                that are missing from the output selection and that will
                be replaced with masked spectra.
            nbins (int): (**Optional**) The total number of defined
                bins.  Default is to find the maximum number in the
                unique_bins list.
            unique_bins (array-like): (**Optional**) The indices of the
                bins that have valid spectra.  The length of this array
                should match the selected_bins array.

        Returns:
            numpy.ndarray: A 2D array with a copy of the data from the
            selected extension.

        Raises:
            KeyError : Raised if `ext` is not a valid extension.
        """

        # Make sure extensions are not lists
        _ext = ext
        if isinstance(ext, list):
            if len(ext) > 1:
                raise ValueError('Must provide only one extension.')
            _ext = ext[0]
        _mask_ext = mask_ext
        if isinstance(_mask_ext, list):
            if len(_mask_ext) > 1:
                raise ValueError('Must provide only one mask extension.')
            _mask_ext = mask_ext[0]

        # Make sure operation is allowed
        if allowed_ext is not None and _ext not in allowed_ext:
            raise KeyError('Cannot copy data to array for extension {0}.'.format(_ext))
        if allowed_ext is not None and _mask_ext is not None and _mask_ext not in allowed_ext:
            raise KeyError('Cannot use extension {0} as mask.'.format(_mask_ext))

        # Make sure masks and flags make sense
        if flag is not None and bitmask is None:
            raise ValueError('If flags are provided, must also provided associated BitMask.')
        if bitmask is not None and not isinstance(bitmask, BitMask):
            raise TypeError('Provided object must be a BitMask.')

        # Total number of available spectra
        nspec = numpy.prod(hdu[_ext].data.shape[0:-1])

        # Create the wavelength mask (will be all true if
        # waverange=None)
        mask = SpectralPixelMask(waverange=waverange).boolean(hdu['WAVE'].data, nspec=nspec)

        # Add in any bitmasks
        if _mask_ext is not None and bitmask is not None:
            mask |= bitmask.flagged(hdu[_mask_ext].data.reshape(nspec,-1), flag=flag)
        if _mask_ext is not None and bitmask is None:
            mask |= (hdu[_mask_ext].data > 0)

        # Create the output MaskedArray
        a = numpy.ma.MaskedArray(hdu[_ext].data.copy().reshape(nspec,-1), mask=mask)

        # Apply any bin selection
        if select_bins is not None:
            a = a[select_bins,:]

        # No missing bins are specified, so return the array
        if missing_bins is None or len(missing_bins) == 0:
            return a

        # Missing bins have been specified, so the unique_bins also need
        # to be specified
        if unique_bins is None:
            raise ValueError('Must specify the unique bins if missing bins are specified.')
        if len(unique_bins) != a.shape[0]:
            raise ValueError('Number of unique bins does not match the number of spectra.')

        # If the number of bins hasn't been provided, construct it based
        # on the unique bins specified
        _nbins = max(unique_bins)+1 if nbins is None else nbins

        # Create the array with the right shape and mask everything
        _a = numpy.ma.MaskedArray(numpy.zeros((_nbins, a.shape[-1]), dtype=hdu[_ext].data.dtype),
                                  mask=numpy.ones((_nbins, a.shape[-1]), dtype=bool))
        # Copy the data to the correct bins, which also unmasks these
        # pixels
        _a[unique_bins,:] = a
        # Return the masked array; "missing bins" are fully masked
        return _a

    @staticmethod
    def marginalize_mask(mask, inp_flags=None, inp_bitmask=None, out_flag=None, out_bitmask=None,
                         out_mask=None, dispaxis=2):
        """
        Marginalize a mask over one of its spectral dimension.

        The output mask selects pixels that are masked by *all* pixels along the
        spectral axis.

        If no optional arguments are provided, the input mask must be a boolean
        array and the result is identically ``numpy.all(mask, axis=dispaxis)``.

        Args:
            mask (`numpy.ndarray`_):
                The mask with more than 1 dimension.  The spectral axis is
                defined by ``dispaxis``.  The array can be either a bitmask or a
                boolean mask; the latter is assumed if ``inp_bitmask`` is not
                provided, and vice versa.
            inp_flags (:obj:`str`, :obj:`list`, optional):
                The list of input flags used to select masked values in bitmask
                arrays.  If None and ``inp_bitmask`` is provided, all flags are
                check individually.
            inp_bitmask (:class:`~mangadap.util.bitmask.BitMask`, optional):
                Object used to identify bits in the input mask.
            out_flag (:obj:`str`, optional):
                If returning a bitmask, this sets the flag to use to identify
                the masked pixels.  If None and ``inp_bitmask`` is provided, the
                output flags are identical to the input flags.
            out_bitmask (:class:`~mangadap.util.bitmask.BitMask`, optional):
                Object used to identify bits in the output mask.
            out_mask (`numpy.ndarray`_, optional):
                If provided, any masking is applied directly to this array;
                i.e., this is both modified and returned by the function.  If
                provided, the array must have the correct shape (same as the
                input array wihtout the spectral axis).
            dispaxis (:obj:`int`, optional):
                The spectral axis of the array.

        Returns:
            `numpy.ndarray`_: A mask array with the marginalized mask.  The
            shape is the same as the input array, but without the spectral axis.

        Raises:
            ValueError:
                Raised if the input arrays have incorrect shapes.
        """
        # Check the input
        if not mask.ndim > 1:
            raise ValueError('Input mask array must have more than one dimension.')

        itype = bool if inp_bitmask is None else inp_bitmask.minimum_dtype()
        # TODO: Need to use 'same_kind' here instead of 'equiv' because MaNGA
        # DRP doesn't necessarily use the minimum dtype for the masks.
        if not numpy.can_cast(mask.dtype, numpy.dtype(itype), casting='same_kind'):
            raise TypeError(f'Cannot cast input mask with type {mask.dtype} to expected type '
                            f'{numpy.dtype(itype)}.')

        spatial_shape = DAPFitsUtil.get_spatial_shape(mask.shape, dispaxis)

        if out_mask is not None and out_mask.shape != spatial_shape:
            raise ValueError('Provided output mask has incorrect shape; expected '
                             f'{spatial_shape}, found {out_mask.shape}.')

        if inp_bitmask is None and inp_flags is not None:
            raise ValueError('Cannot interpret input flags without inp_bitmask.')
        if out_bitmask is None and out_flag is not None:
            raise ValueError('Cannot interpret output flags without out_bitmask.')
        if out_bitmask is not None and out_flag is None:
            raise ValueError('If providing the output bitmask, must also provide the output flag.')

        otype = bool if out_bitmask is None else out_bitmask.minimum_dtype()
        # TODO: May need to use 'same_kind' instead of 'equiv'
        if out_mask is not None \
                and not numpy.can_cast(out_mask.dtype, numpy.dtype(otype), casting='equiv'):
            raise TypeError('Provided output mask has incorrect type; expected '
                            f'{numpy.dtype(otype)}, found {out_mask.dtype}.')
        _out_mask = numpy.zeros(spatial_shape, dtype=otype) if out_mask is None else out_mask

        if inp_bitmask is None:
            # Handle the case when the input mask is a boolean array
            indx = numpy.all(mask, axis=dispaxis)
            if out_bitmask is None:
                _out_mask |= indx
                return _out_mask
            _out_mask[indx] = out_bitmask.turn_on(_out_mask[indx], out_flag)
            return _out_mask

        # Handle the case when the input mask is a bit mask
        if inp_flags is None:
            inp_flags = inp_bitmask.keys()
        _inp_flags = inp_flags if isinstance(inp_flags, list) else [inp_flags]

        for flag in _inp_flags:
            indx = numpy.all(inp_bitmask.flagged(mask, flag=flag), axis=dispaxis)
            if numpy.sum(indx) == 0:
                continue
            if out_bitmask is None:
                _out_mask |= indx
                continue
            _out_mask[indx] = out_bitmask.turn_on(_out_mask[indx],
                                                  (flag if out_flag is None else out_flag))
        return _out_mask


    @staticmethod
    def _mask_data_type(bit_type):
        if bit_type in [numpy.uint64, numpy.int64]:
            return ('FLAG64BIT', '64-bit mask')
        if bit_type in [numpy.uint32, numpy.int32]:
            return ('FLAG32BIT', '32-bit mask')
        if bit_type in [numpy.uint16, numpy.int16]:
            return ('FLAG16BIT', '16-bit mask')
        if bit_type in [numpy.uint8, numpy.int8]:
            return ('FLAG8BIT', '8-bit mask')
        if bit_type in [bool, numpy.bool_]:
            return ('MASKZERO', 'Binary mask; zero values are good/unmasked')
        raise ValueError('Invalid bit_type: {0}!'.format(str(bit_type)))


    @staticmethod
    def empty_hdus(ext):
        hdus = []
        for e in ext:
            hdus += [ fits.ImageHDU(data=None, name=e) ]
        return hdus


    @staticmethod
    def list_of_image_hdus(data, hdr, ext):
        hdus = []
        for d,h,e in zip(data,hdr,ext):
            hdus += [ fits.ImageHDU(data=d, header=h, name=e) ]
        return hdus


    @staticmethod
    def redshift_to_Newtonian_velocity(v, redshift, ivar=False):
        if ivar:
            return v*numpy.square(1.0+redshift)
        return (v-astropy.constants.c.to('km/s').value*redshift)/(1.0+redshift)





# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
A class hierarchy that performs the stellar-continuum fitting.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import inspect
from pathlib import Path
import warnings
import logging

from IPython import embed

import numpy

from astropy.io import fits
import astropy.constants

from ..par.parset import KeywordParSet, ParSet
from ..util.log import log_output
from ..util.fitsutil import DAPFitsUtil
from ..util.fileio import create_symlink
from ..util.sampling import spectral_coordinate_step, spectrum_velocity_scale
from ..util.resolution import SpectralResolution
from ..util.dapbitmask import DAPBitMask
from .spatiallybinnedspectra import SpatiallyBinnedSpectra
from .templatelibrary import TemplateLibrary
from .ppxffit import PPXFFitPar, PPXFFit
from .util import replace_with_data_from_nearest_coo


class StellarContinuumModelDef(KeywordParSet):
    """
    A class that holds the parameters necessary to perform the
    stellar-continuum fitting.

    The provided ``fitfunc`` must have the form::

        model_wave, model_flux, model_mask, model_par \
                = fitfunc(self, binned_spectra, par=None)

    where ``binned_spectra`` has type
    :class:`mangadap.proc.spatiallybinnedspetra.SpatiallyBinnedSpectra`,
    and the returned objects are the wavelength vector, the fitted
    model flux, a bitmask for the model, and the set of model
    parameters. For example, see
    :func:`mangadap.proc.ppxffit.PPXFFit.fit_SpatiallyBinnedSpectra`.

    The defined parameters are:

    .. include:: ../tables/stellarcontinuummodeldef.rst
    """
    def __init__(self, key='MILESHC', minimum_snr=1.0, fitpar=None, fitclass=None, fitfunc=None,
                 overwrite=False):

        # Use the signature to get the parameters and the default values
        sig = inspect.signature(self.__class__)
        pars = list(sig.parameters.keys())
        defaults = [sig.parameters[key].default for key in pars]

        # Remaining definitions done by hand
        in_fl = [int, float]
        par_opt = [ParSet, dict]

        values = [key, minimum_snr, fitpar, fitclass, fitfunc, overwrite ]
        dtypes = [str, in_fl, par_opt, None, None, bool]
        can_call = [False, False, False, False, True, False]
        descr = ['Keyword used to distinguish between different spatial binning schemes.',
                 'Minimum S/N of spectrum to fit',
                 'Any additional parameters, aside from the spectra themselves, required by ' \
                    'the fitting function.  If None, uses a default pPXF fit.',
                 'Instance of class object to use for the model fitting.  Needed in case ' \
                    'fitfunc is a non-static member function of a class.  If None, uses a ' \
                    'default pPXF fit.',
                 'The function that models the spectra.  If None, uses a default pPXF fit and ' \
                    'anything provided for fitpar and fitclass are ignored!',
                 'If the output file already exists, redo all the calculations and overwrite it.']

        super().__init__(pars, values=values, defaults=defaults, dtypes=dtypes, can_call=can_call,
                         descr=descr)
        self._validate()

    def _validate(self):
        """
        Validate the modeling method.
        """
        if self['fitfunc'] is None:
            warnings.warn('No fitting function defined.  Defaulting to use mangadap.proc.ppxffit.')
            self['fitpar'] = PPXFFitPar()
            self['fitclass'] = PPXFFit(StellarContinuumModelBitMask())
            self['fitfunc'] = self['fitclass'].fit_SpatiallyBinnedSpectra

    @classmethod
    def from_dict(cls, d):
        """
        Instantiate the object from a nested dictionary.

        The dictionary is expected to have keys for all the class keywords; see
        the class instantiation method.  It should also include the
        ``'fit_method'`` keyword used to select the type of fit to perform, and
        a nested dictionary selected by the keyword ``'fit'`` with the 
        parameters required by the fitting method.  If these are not provided,
        the default fitter, :class:`~mangadap.proc.ppxffit.PPXFFit` and
        parameters will be used.

        .. warning::

            Currently, the only accepted value for ``'fit_method'`` is
            ``'ppxf'``, and the ``'fit'`` dictionary must provide the parameters
            defined by :class:`~mangadap.proc.ppxffit.PPXFFitPar`.

        Args:
            d (:obj:`dict`):
                Dictionary used to instantiate the class.
        """
        # Copy over the primary keywords
        _d = {}
        for key in ['key', 'minimum_snr', 'overwrite']:
            if key in d.keys():
                _d[key] = d[key]

        # Set the spatial binning method
        if 'fit_method' not in d or d['fit_method'] in ['none', None]:
            # Use the default fitter and default fitting parameters!
            _d['fitpar'] = None
            _d['fitclass'] = None
            _d['fitfunc'] = None
        elif d['fit_method'] == 'ppxf':
            _d['fitpar'] = PPXFFitPar.from_dict(d['fit']) if 'fit' in d else PPXFFitPar()
            _d['fitclass'] = PPXFFit(StellarContinuumModelBitMask())
            _d['fitfunc'] = _d['fitclass'].fit_SpatiallyBinnedSpectra
        else:
            raise ValueError('Unrecognized fitting method.  Requires direct instantiation or '
                             'code alterations.')
        # Return the instantiation
        return super().from_dict(_d)


class StellarContinuumModelBitMask(DAPBitMask):
    r"""
    Derived class that specifies the mask bits for the stellar-continuum
    modeling. The maskbits defined are:

    .. include:: ../tables/stellarcontinuummodelbitmask.rst
    """
    cfg_root = 'stellar_continuum_model_bits'


class StellarContinuumModel:
    r"""
    Class that holds the stellar-continuum model results.

    .. todo::

        - Update attributes

    Args:
        method (:class:`~mangadap.proc.stellarcontinuummodel.StellarContinuumModelDef`):
            Object that defines the main parameters used for the stellar
            continuum fitting.
        binned_spectra (:class:`mangadap.proc.spatiallybinnedspectra.SpatiallyBinnedSpectra`):
            The binned spectra to fit. Must be able to access parent
            :class:`mangadap.datacube.datacube.DataCube` attribute
            (``cube``).
        guess_vel (:obj:`float`, `numpy.ndarray`_):
            A single or spectrum-dependent initial estimate of the
            redshift in km/s (:math:`cz`).
        guess_sig (:obj:`float`, `numpy.ndarray`_, optional):
            A single or spectrum-dependent initial estimate of the
            velocity dispersion in km/s. If None, default set to 100
            km/s.
        method_list (:obj:`list`, optional):
            List of :class:`StellarContinuumModelDef` objects that
            define one or more methods to use for the stellar
            continuum fitting. The default list is provided by the
            config files in the DAP source directory and compiled
            into this list using
            :func:`available_stellar_continuum_modeling_methods`.
        output_path (:obj:`str`, `Path`_, optional):
            The path for the output file.  If None, the current working
            directory is used.
        output_file (:obj:`str`, optional):
            The name of the output "reference" file. The full path of the output
            file will be :attr:`directory_path`/:attr:`output_file`.  If None,
            the default is to combine ``cube.output_root`` and the method keys.
            The order of the keys is the order of operations (rdxqa, binning,
            stellar continuum).  See
            :func:`~mangadap.proc.stellarcontinuummodel.StellarContinuumModel.default_paths`.
        hardcopy (:obj:`bool`, optional):
            Flag to write the HDUList attribute to disk. If False,
            the core `astropy.io.fits.HDUList`_ attribute
            (:attr:`hdu`) is only kept in memory and would have to be
            reconstructed.
        symlink_dir (:obj:`str`, optional):
            Create a symbolic link to the created file in the supplied
            directory.  If None, no symbolic link is created.
        tpl_hardcopy (:obj:`bool`, optional):
            Save the processed template library used during the fit.  If True,
            the output path and symlink directory for the template library are
            identical to the main reference file.
        overwrite (:obj:`bool`, optional):
            Overwrite any existing files. Default is to use any
            existing file instead of redoing the analysis and
            overwriting the existing output.
        checksum (:obj:`bool`, optional):
            Use the checksum in the fits header to confirm that the
            data has not been corrupted. The checksum is **always**
            written to the fits header when the file is created.
        loggers (:obj:`list`, optional):
            List of `logging.Logger`_ objects to log progress;
            ignored if quiet=True. Logging is done using
            :func:`mangadap.util.log.log_output`. Default is no
            logging.
        quiet (:obj:`bool`, optional):
            Suppress all terminal and logging output.

    Attributes:
        loggers (:obj:`list`):
            List of `logging.Logger`_ objects to log progress.
        quiet (:obj:`bool`):
            Suppress all terminal and logging output.

    """
    def __init__(self, method, binned_spectra, guess_vel, guess_sig=None, method_list=None,
                 output_path=None, output_file=None, hardcopy=True, symlink_dir=None,
                 tpl_hardcopy=False, overwrite=False, checksum=False, loggers=None, quiet=False):

        self.loggers = None
        self.quiet = False

        # Define the method properties
        self.method = method
        if not isinstance(self.method, StellarContinuumModelDef):
            raise TypeError('Method must have type StellarContinuumModelDef.')

        self.binned_spectra = None
        self.guess_vel = None
        self.guess_sig = None

        # Define the output directory and file
        self.directory_path = None      # Set in default_paths
        self.output_file = None
        self.hardcopy = None
        self.tpl_hardcopy = None
        self.symlink_dir = None

        # Define the bitmask
        self.bitmask = StellarContinuumModelBitMask()

        # Initialize the main class attributes
        self.hdu = None
        self.checksum = checksum
        self.shape = None
        self.spatial_shape = None
        self.nspec = None
        self.spatial_index = None
        self.spectral_arrays = None
        self._assign_spectral_arrays()
        self.image_arrays = None
        self._assign_image_arrays()
        self.nwave = None

        self.nmodels = None
        self.missing_models = None

        # TODO: Include covariance between measured properties

        # Run the assessments of the DRP file
        self.fit(binned_spectra, guess_vel, guess_sig=guess_sig, output_path=output_path,
                 output_file=output_file, hardcopy=hardcopy, symlink_dir=symlink_dir,
                 tpl_hardcopy=tpl_hardcopy, overwrite=overwrite, loggers=loggers, quiet=quiet)

    def __getitem__(self, key):
        return self.hdu[key]

    def get_template_library(self, output_path=None, output_file=None, hardcopy=None,
                             symlink_dir=None, velocity_offset=None, match_resolution=False,
                             resolution_fwhm=None):
        """
        Construct the template library for the continuum fitting.

        If the FWHM (in angstroms; see ``resolution_fwhm``) is not
        provided, this method is a simple wrapper that returns a
        :class:`mangadap.proc.templatelibrary.TemplateLibrary`.
        Otherwise, the provided FWHM is used to construct the
        spectral resolution for the template library.

        In either case, the sampling is set to match the spectra in
        :attr:`binned_spectra` (up to the velocity-scale ratio set by
        :attr:`method`).

        Args:
            output_path (:obj:`str`, `Path`_, optional):
                The path for the output file.  If None, set to 
                :attr:`directory_path`.  An output file is only produced if
                ``hardcopy`` is True.
            output_file (:obj:`str`, optional):
                The name of the output template library file.  If None and
                ``resolution_fwhm`` is None, uses the default template library
                file; see
                :func:`~mangadap.proc.templatelibrary.TemplateLibrary.default_paths`.
                If None and ``resolution_fwhm`` is provided, the default
                template file is also used, but the library key is changed to
                indicate the new FHWM of the output resolution.  An output file
                is only produced if ``hardcopy`` is True.
            hardcopy (:obj:`bool`, optional):
                Flag to keep a hardcopy of the processed template library.  If
                None, the flag is set by :attr:`tpl_hardcopy`.  If no hardcopy
                is to be produced, ``output_path`` and ``output_file`` are
                ignored.
            symlink_dir (:obj:`str`, optional):
                Create a symbolic link to the created template library file in
                this directory. If None, :attr:`symlink_dir` is used.  A symlink
                is produced only if ``hardcopy`` is True.
            velocity_offset (:obj:`float`, optional):
                Velocity offset to use when matching the spectral
                resolution between the template library and the galaxy
                spectra.
            match_resolution (:obj:`bool`, optional):
                Match the spectral resolution of the template library to
                the resolution provided by the ``cube``; the latter must
                be provided for this argument to have any use.
            resolution_fwhm (:obj:`float`, optional):
                The target resolution for the template library set by
                a constant FWHM of the resolution element in
                angstroms over all wavelengths. If None, the
                resolution is either matched to the binned spectra
                (see ``match_resolution``) or kept at the native
                resolution of the library.

        Returns:
            :class:`mangadap.proc.templatelibrary.TemplateLibrary`:
            The template library prepared for fitting the binned
            spectra.
        """

        # Allow output paths and files to be over-ridden
        _output_path = self.directory_path if output_path is None else output_path
        _hardcopy = self.tpl_hardcopy if hardcopy is None else hardcopy
        _symlink_dir = self.symlink_dir if symlink_dir is None else symlink_dir

        if resolution_fwhm is None:
            return TemplateLibrary(self.method['fitpar']['template_library_key'],
                                   velocity_offset=velocity_offset, cube=self.binned_spectra.cube,
                                   match_resolution=match_resolution,
                                   velscale_ratio=self.method['fitpar']['velscale_ratio'],
                                   output_path=_output_path, output_file=output_file,
                                   hardcopy=_hardcopy, symlink_dir=_symlink_dir,
                                   loggers=self.loggers, quiet=self.quiet)

        if output_file is None:
            _key = f'{self.method["fitpar"]["template_library_key"]}-FWHM{resolution_fwhm:.2f}'
            output_file = TemplateLibrary.default_paths(_key, cube=self.binned_spectra.cube,
                                                        output_path=_output_path)[1]

        # Set the spectral resolution
        wave = self.binned_spectra['WAVE'].data
        sres = SpectralResolution(wave, wave/resolution_fwhm, log10=True)

        velscale_ratio = 1 if self.method['fitpar']['velscale_ratio'] is None \
                            else self.method['fitpar']['velscale_ratio']
        spectral_step = spectral_coordinate_step(wave, log=True) / velscale_ratio

        return TemplateLibrary(self.method['fitpar']['template_library_key'], sres=sres,
                                velocity_offset=velocity_offset, spectral_step=spectral_step,
                                log=True, output_path=_output_path, output_file=output_file,
                                hardcopy=_hardcopy, symlink_dir=_symlink_dir,
                                loggers=self.loggers, quiet=self.quiet)

    def _initialize_primary_header(self, hdr=None):
        """
        Construct the primary header for the reference file.

        Args:
            hdr (`astropy.io.fits.Header`_, optional):
                Input base header for added keywords. If None, uses
                the datacube header from :attr:`binned_spectra` (if
                there is one) and then cleans the header using
                :func:`mangadap.util.fitsutil.DAPFitsUtil.clean_dap_primary_header`.

        Returns:
            `astropy.io.fits.Header`_: Initialized header object.
        """
        # Copy the from the DRP and clean it
        if hdr is None:
            hdr = self.binned_spectra.cube.prihdr.copy()
            hdr = DAPFitsUtil.clean_dap_primary_header(hdr)
        
        # Add keywords specific to this object
        hdr['AUTHOR'] = 'Kyle B. Westfall <westfall@ucolick.org>'
        hdr['VSTEP'] = (spectrum_velocity_scale(self.binned_spectra['WAVE'].data),
                        'Velocity step per spectral channel.')
        hdr['SCKEY'] = (self.method['key'], 'Stellar-continuum modeling method keyword')
        hdr['SCMINSN'] = (self.method['minimum_snr'], 'Minimum S/N of spectrum to include')

        # Guess kinematics
        # TODO: These are currently single numbers, but they may
        # eventually be vectors!  Include guess kinematics as a map?
        if self.guess_vel is not None:
            hdr['SCINPVEL'] = (self.guess_vel, 'Initial guess velocity')
        if self.guess_sig is not None:
            hdr['SCINPSIG'] = (self.guess_sig, 'Initial guess velocity dispersion')

        # Include the number of models
        hdr['NSCMOD'] = (self.nmodels, 'Number of unique stellar-continuum models')
        return hdr

    def _add_method_header(self, hdr):
        """
        Add method-specific metadata to the header.

        Note that the header object is both edited in-place and
        returned.

        Args:
            hdr (`astropy.io.fits.Header`_):
                Input base header for added keywords. Expected to
                have been initialized using
                :func:`_initialize_primary_header`.

        Returns:
            `astropy.io.fits.Header`_: Edited header object.
        """
        if self.method['fitclass'] is not None:
            try:
                hdr['SCTYPE'] = self.method['fitclass'].fit_type
                hdr['SCMETH'] = self.method['fitclass'].fit_method
            except:
                if not self.quiet and self.hardcopy:
                    warnings.warn('Fit class object does not have fit_type and/or fit_method ' \
                                  'attributes.  No parameters written to header.')
        if self.method['fitpar'] is not None:
            try:
                hdr = self.method['fitpar'].toheader(hdr)
            except:
                if not self.quiet and self.hardcopy:
                    warnings.warn('Fit parameter class has no toheader() function.  No ' \
                                  'parameters written to header.')
        return hdr


    def _finalize_cube_mask(self, mask):
        """
        Finalize the mask after the 2D mask has been reconstructed
        into a 3D cube.

        This mostly handles the masks for regions outside the
        datacube field of view.

        Note that the input mask is both edited in-place and
        returned.

        .. todo::

            - This needs to be abstracted for non-DRP datacubes.
            - Describe MAPMASK usage

        Args:
            mask (`numpy.ndarray`_):
                3D array with the current bitmask data.

        Returns:
            `numpy.ndarray`_: Edited bitmask data.
        """
        # TODO: I'm not sure this is correct for non-MaNGA data
        # Turn on the flag stating that the pixel wasn't used
        indx = self.binned_spectra.bitmask.flagged(self.binned_spectra.cube.mask,
                                                   flag=self.binned_spectra.do_not_fit_flags())
        mask[indx] = self.bitmask.turn_on(mask[indx], 'DIDNOTUSE')

        # Turn on the flag stating that the pixel has a foreground star
        indx = self.binned_spectra.bitmask.flagged(self.binned_spectra.cube.mask, flag='FORESTAR')
        mask[indx] = self.bitmask.turn_on(mask[indx], 'FORESTAR')

        # Propagate the MAPMASK to the full cube
        for f in ['DIDNOTUSE', 'FORESTAR', 'LOW_SNR']:
            indx = self.bitmask.flagged(self['MAPMASK'].data, f)
            mask[indx,:] = self.bitmask.turn_on(mask[indx,:], f)

        return mask

    def _assign_spectral_arrays(self):
        """
        Set :attr:`spectral_arrays`, which contains the list of
        extensions in :attr:`hdu` that contain spectral data.
        """
        self.spectral_arrays = ['FLUX', 'MASK']

    def _assign_image_arrays(self):
        """
        Set :attr:`image_arrays`, which contains the list of extensions
        in :attr:`hdu` that are on-sky image data.
        """
        self.image_arrays = ['BINID']

    def _get_missing_models(self, debug=False):
        """
        Find the spectra with missing models, either because there is
        no model or the bin is also missing.

        Args:
            debug (:obj:`bool`, optional):
                When checking the S/N limits, run
                :func:`mangadap.proc.spatiallybinnedspectra.SpatiallyBinnedSpectra.above_snr_limit`
                in debug mode.

        Returns:
            `numpy.ndarray`_: Sorted list of missing models.
        """
        good_snr = self.binned_spectra.above_snr_limit(self.method['minimum_snr'], debug=debug)
        return numpy.sort(self.binned_spectra['BINS'].data['BINID'][numpy.invert(good_snr)].tolist()
                                + self.binned_spectra.missing_bins) 

    def _construct_2d_hdu(self, good_snr, model_flux, model_mask, model_par):
        r"""
        Construct :attr:`hdu` that is held in memory for manipulation
        of the object. See :func:`construct_3d_hdu` to convert the
        object into a datacube.

        Args:
            good_snr (`numpy.ndarray`_):
                Boolean array selecting the bins with good S/N.
            model_flux (`numpy.ndarray`_):
                Best-fitting model spectra for each bin.
            model_mask (`numpy.ndarray`_):
                Bitmask flags for each model pixel.
            model_par (:class`~mangadap.proc.spectralfitting.StellarKinematicsFitDataTable`):
                Data table with the best-fitting model parameters.
        """
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, 'Constructing hdu ...')

        # Initialize the headers
        pri_hdr = self._initialize_primary_header()
        pri_hdr = self._add_method_header(pri_hdr)
        map_hdr = DAPFitsUtil.build_map_header(self.binned_spectra.cube.fluxhdr,
                                               'K Westfall <westfall@ucolick.org>')

        # Get the spatial map mask
        map_mask = numpy.zeros(self.spatial_shape, dtype=self.bitmask.minimum_dtype())
        # Add any spaxel not used because it was flagged by the binning
        # step
        indx = self.binned_spectra['MAPMASK'].data > 0
        map_mask[indx] = self.bitmask.turn_on(map_mask[indx], 'DIDNOTUSE')
        # Isolate any spaxels with foreground stars
        indx = self.binned_spectra.bitmask.flagged(self.binned_spectra['MAPMASK'].data, 'FORESTAR')
        map_mask[indx] = self.bitmask.turn_on(map_mask[indx], 'FORESTAR')
        # Get the bins that were blow the S/N limit
        indx = numpy.invert(DAPFitsUtil.reconstruct_map(self.spatial_shape,
                                                        self.binned_spectra['BINID'].data.ravel(),
                                                        good_snr, dtype='bool')) & (map_mask == 0)
        map_mask[indx] = self.bitmask.turn_on(map_mask[indx], 'LOW_SNR')

        # Get the bin ids with fitted models
        bin_indx = DAPFitsUtil.downselect_bins(self.binned_spectra['BINID'].data.ravel(),
                                               model_par['BINID']).reshape(self.spatial_shape)

        # Save the data to the hdu attribute
        self.hdu = fits.HDUList([fits.PrimaryHDU(header=pri_hdr),
                                 fits.ImageHDU(data=model_flux.data, name='FLUX'),
                                 fits.ImageHDU(data=model_mask, name='MASK'),
                                 self.binned_spectra['WAVE'].copy(),
                                 fits.ImageHDU(data=bin_indx, header=map_hdr, name='BINID'),
                                 fits.ImageHDU(data=map_mask, header=map_hdr, name='MAPMASK'),
                                 model_par.to_hdu(name='PAR')])

    @staticmethod
    def default_paths(cube, method_key, rdxqa_method, binning_method, output_path=None,
                      output_file=None):
        """
        Set the default directory and file name for the output file.

        Args:
            cube (:class:`mangadap.datacube.datacube.DataCube`):
                Datacube to analyze.
            method_key (:obj:`str`):
                Keyword designating the method used for the reduction
                assessments.
            rdxqa_method (:obj:`str`):
                The method key for the basic assessments of the datacube.
            binning_method (:obj:`str`):
                The method key for the spatial binning.
            output_path (:obj:`str`, `Path`_, optional):
                The path for the output file.  If None, the current working
                directory is used.
            output_file (:obj:`str`, optional):
                The name of the output "reference" file. The full path of the
                output file will be :attr:`directory_path`/:attr:`output_file`.
                If None, the default is to combine ``cube.output_root`` and the
                method keys.  The order of the keys is the order of operations
                (rdxqa, binning, stellar continuum).

        Returns:
            :obj:`tuple`: Returns a `Path`_ with the output directory and a
            :obj:`str` with the output file name.
        """
        directory_path = Path('.').resolve() if output_path is None \
                                else Path(output_path).resolve()
        method = f'{rdxqa_method}-{binning_method}-{method_key}'
        _output_file = f'{cube.output_root}-{method}.fits.gz' if output_file is None \
                            else output_file
        return directory_path, _output_file

    def file_name(self):
        """Return the name of the output file."""
        return self.output_file

    def file_path(self):
        """Return the full path to the output file."""
        if self.directory_path is None or self.output_file is None:
            return None
        return self.directory_path / self.output_file
    
    def info(self):
        return self.hdu.info()

    def all_spectrum_flags(self):
        return ['DIDNOTUSE', 'FORESTAR', 'LOW_SNR', 'ARTIFACT', 'OUTSIDE_RANGE', 'EML_REGION',
                'TPL_PIXELS', 'TRUNCATED', 'PPXF_REJECT', 'INVALID_ERROR', 'FIT_FAILED',
                'NEAR_BOUND']

    def all_except_emission_flags(self):
        return ['DIDNOTUSE', 'FORESTAR', 'LOW_SNR', 'ARTIFACT', 'OUTSIDE_RANGE', 'TPL_PIXELS',
                'TRUNCATED', 'PPXF_REJECT', 'INVALID_ERROR', 'FIT_FAILED', 'NEAR_BOUND']

    def fit(self, binned_spectra, guess_vel, guess_sig=None, output_path=None, output_file=None,
            hardcopy=True, symlink_dir=None, tpl_hardcopy=None, overwrite=None, loggers=None,
            quiet=False):
        """
        Fit the binned spectra using the specified method.

        Args:
            binned_spectra (:class:`mangadap.proc.spatiallybinnedspectra.SpatiallyBinnedSpectra`):
                The binned spectra to fit. Must be able to access
                parent :class:`mangadap.datacube.datacube.DataCube`
                attribute (``cube``).
            guess_vel (:obj:`float`, `numpy.ndarray`):
                A single or spectrum-dependent initial estimate of the
                redshift in km/s (:math:`cz`).
            guess_sig (:obj:`float`, `numpy.ndarray`, optional):
                A single or spectrum-dependent initial estimate of the
                velocity dispersion in km/s.
            output_path (:obj:`str`, `Path`_, optional):
                The path for the output file.  If None, the current working
                directory is used.
            output_file (:obj:`str`, optional):
                The name of the output "reference" file. The full path of the output
                file will be :attr:`directory_path`/:attr:`output_file`.  If None,
                the default is to combine ``cube.output_root`` and the method keys.
                The order of the keys is the order of operations (rdxqa, binning,
                stellar continuum).  See
                :func:`~mangadap.proc.stellarcontinuummodel.StellarContinuumModel.default_paths`.
            hardcopy (:obj:`bool`, optional):
                Flag to write the HDUList attribute to disk.  Default is
                True; if False, the HDUList is only kept in memory and
                would have to be reconstructed.
            symlink_dir (:obj:`str`, optional):
                Create a symbolic link to the created file in the
                supplied directory.  Default is to produce no symbolic
                link. 
            tpl_hardcopy (:obj:`bool`, optional):
                Save the processed template library used during the fit.  If
                True, the output path and symlink directory for the template
                library are identical to the main reference file.
            overwrite (:obj:`bool`, optional):
                Overwrite any existing files.  Default is to use any
                existing file instead of redoing the analysis and
                overwriting the existing output.
            loggers (:obj:`list`, optional):
                List of `logging.Logger`_ objects to log progress;
                ignored if quiet=True.  Logging is done using
                :func:`mangadap.util.log.log_output`.  Default is no
                logging.
            quiet (:obj:`bool`, optional):
                Suppress all terminal and logging output. 
        """
        # Initialize the reporting
        if loggers is not None:
            self.loggers = loggers
        self.quiet = quiet

        # SpatiallyBinnedSpectra object always needed
        if binned_spectra is None:
            raise ValueError('Must provide spectra object for fitting.')
        if not isinstance(binned_spectra, SpatiallyBinnedSpectra):
            raise TypeError('Must provide a valid SpatiallyBinnedSpectra object!')
        if binned_spectra.hdu is None:
            raise ValueError('Provided SpatiallyBinnedSpectra object is undefined!')
        self.binned_spectra = binned_spectra

        self.shape = self.binned_spectra.shape
        self.spatial_shape =self.binned_spectra.spatial_shape
        self.nspec = self.binned_spectra.nspec
        self.spatial_index = self.binned_spectra.spatial_index.copy()
        self.nwave = self.binned_spectra.nwave
        
        # Get the guess kinematics
        if guess_vel is not None:
            self.guess_vel=guess_vel
        if self.guess_vel is None:
            raise ValueError('Must provide initial guess velocity.')
        self.guess_sig = 100.0 if guess_sig is None else guess_sig

        #---------------------------------------------------------------
        # Get the good spectra
#        debug = True
        debug = False
        good_snr = self.binned_spectra.above_snr_limit(self.method['minimum_snr'], debug=debug)

        # Report
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, '-'*50)
            log_output(loggers, 1, logging.INFO, f'{"STELLAR CONTINUUM FITTING":^50}')
            log_output(self.loggers, 1, logging.INFO, '-'*50)
            log_output(self.loggers, 1, logging.INFO,
                       f'Number of binned spectra: {self.binned_spectra.nbins}')
            if len(self.binned_spectra.missing_bins) > 0:
                log_output(self.loggers, 1, logging.INFO,
                           f'Missing bins: {len(self.binned_spectra.missing_bins)}')
            log_output(self.loggers, 1, logging.INFO,
                       f'With good S/N and to fit: {numpy.sum(good_snr)}')

        # TODO: Is there a better way to handle this?
        if numpy.sum(good_snr) == 0:
            raise ValueError('No good spectra to fit!')

        #---------------------------------------------------------------
        # Fill in any remaining binning parameters
        self.tpl_hardcopy = tpl_hardcopy
#        self._fill_method_par()
        if self.method['fitpar'] is not None and hasattr(self.method['fitpar'], 'fill'):
            self.method['fitpar'].fill(self.binned_spectra, guess_vel=self.guess_vel,
                                       guess_sig=self.guess_sig, output_path=self.directory_path,
                                       hardcopy=self.tpl_hardcopy, symlink_dir=self.symlink_dir,
                                       loggers=self.loggers, quiet=self.quiet)

        _overwrite = self.method['overwrite'] if overwrite is None else overwrite

        # (Re)Set the output paths
        self.directory_path, self.output_file \
                = StellarContinuumModel.default_paths(self.binned_spectra.cube,
                                                      self.method['key'],
                                                      self.binned_spectra.rdxqa.method['key'],
                                                      self.binned_spectra.method['key'],
                                                      output_path=output_path,
                                                      output_file=output_file)

        #---------------------------------------------------------------
        # Check that the file path is defined
        ofile = self.file_path()
        if ofile is None:
            raise ValueError('File path for output file is undefined!')

        # Report
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, f'Output path: {self.directory_path}')
            log_output(self.loggers, 1, logging.INFO, f'Output file: {self.output_file}')
        
        #---------------------------------------------------------------
        # If the file already exists, and not ovewriting, just read the
        # file
        self.symlink_dir = symlink_dir
        if ofile.exists() and not overwrite:
            self.hardcopy = True
            if not self.quiet:
                log_output(self.loggers, 1, logging.INFO, 'Using existing file')
            self.read(checksum=self.checksum, debug=debug)
            # Make sure the symlink exists
            if self.symlink_dir is not None:
                create_symlink(str(ofile), self.symlink_dir, overwrite=overwrite)
            if not self.quiet:
                log_output(self.loggers, 1, logging.INFO, '-'*50)
            return

        #---------------------------------------------------------------
        # Fit the spectra
        # Mask should be fully defined within the fitting function
        model_wave, model_flux, model_mask, model_par \
                = self.method['fitfunc'](self.binned_spectra, gpm=good_snr,
                                         par=self.method['fitpar'], loggers=self.loggers,
                                         quiet=self.quiet, debug=debug)

        # The number of models returned should match the number of
        # "good" spectra

        # TODO: Include failed fits in "missing" models?

        # DEBUG
        if model_flux.shape[0] != numpy.sum(good_snr):
            raise ValueError('Unexpected returned shape of fitted continuum models.')

        # Set the number of models and the missing models
        self.nmodels = model_flux.shape[0]
        self.missing_models = self._get_missing_models(debug=debug)

        # Report
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, f'Fitted models: {self.nmodels}')
            if len(self.missing_models) > 0:
                log_output(self.loggers, 1, logging.INFO,
                           f'Missing models: {len(self.missing_models)}')

        # Construct the 2d hdu list that only contains the fitted models
        self.hardcopy = hardcopy
        self._construct_2d_hdu(good_snr, model_flux, model_mask, model_par)

        #---------------------------------------------------------------
        # Write the data, if requested
        if self.hardcopy:
            if not self.directory_path.exists():
                self.director_path.mkdir(parents=True)
            self.write(overwrite=overwrite)
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, '-'*50)

    def construct_3d_hdu(self):
        """
        Reformat the model spectra into a cube matching the shape of
        the DRP fits file.
        """
        # Report
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, 'Constructing stellar continuum datacube ...')

        bin_indx = self.hdu['BINID'].data.copy()
        model_flux = self.hdu['FLUX'].data.copy()
        model_mask = self.hdu['MASK'].data.copy()

        flux, mask = DAPFitsUtil.reconstruct_cube(self.shape, bin_indx.ravel(),
                                                  [model_flux, model_mask])

        mask = self._finalize_cube_mask(mask)

        # Primary header is identical regardless of the shape of the
        # extensions
        hdr = self.hdu['PRIMARY'].header.copy()
        cube_hdr = DAPFitsUtil.build_cube_header(self.binned_spectra.cube,
                                                 'K Westfall <westfall@ucolick.org>')

        # Return the converted hdu without altering the internal hdu
        return fits.HDUList([ fits.PrimaryHDU(header=hdr),
                              fits.ImageHDU(data=flux, header=cube_hdr, name='FLUX'),
                              fits.ImageHDU(data=mask, header=cube_hdr, name='MASK'),
                              self.hdu['WAVE'].copy(),
                              self.hdu['BINID'].copy(),
                              self.hdu['PAR'].copy()
                            ])

    # Exact same function as used by SpatiallyBinnedSpectra
    def write(self, match_datacube=False, overwrite=False):
        """
        Write the hdu object to the file.

        Args:
            match_datacube (:obj:`bool`, optional):
                Match the shape of the data arrays to the input
                datacube. I.e., convert them to 3D and replicate the
                binned spectrum to each spaxel in the bin.
            overwrite (:obj:`bool`, optional):
                Overwrite any existing file.
        """
        # Convert the spectral arrays in the HDU to a 3D cube and write
        # it
        if match_datacube:
            hdu = self.construct_3d_hdu()
            DAPFitsUtil.write(hdu, str(self.file_path()), overwrite=overwrite, checksum=True,
                              symlink_dir=self.symlink_dir, loggers=self.loggers, quiet=self.quiet)
            return
        # Just write the unique (2D) data
        DAPFitsUtil.write(self.hdu, str(self.file_path()), overwrite=overwrite, checksum=True,
                          symlink_dir=self.symlink_dir, loggers=self.loggers, quiet=self.quiet) 

    def read(self, ifile=None, strict=True, checksum=False, debug=False):
        """
        Read an existing file with a previously fit set of continuum
        models.

        Args:
            ifile (:obj:`str`, optional):
                Name of the file with the data. Default is to use the
                name provided by :func:`file_path`.
            strict (:obj:`bool`, optional):
                Force a strict reading of the file to make sure that
                it adheres to the expected format. At the moment,
                this only checks to make sure the method keyword
                printed in the header matches the expected value in
                :attr:`method`.
            checksum (:obj:`bool`, optional):
                Use the checksum in the header to check for
                corruption of the data.
            debug (:obj:`bool`, optional):
                Run check for missing models in debugging mode.

        Raises:
            FileNotFoundError:
                Raised if the file does not exist.
            ValueError:
                Raised if ``strict`` is true and the header keyword
                ``SCKEY`` does not match the method keyword.
        """
        if ifile is None:
            ifile = self.file_path()
        if not ifile.exists():
            raise FileNotFoundError('File does not exist!: {0}'.format(ifile))

        if self.hdu is not None:
            self.hdu.close()

#        self.hdu = fits.open(ifile, checksum=checksum)
        self.hdu = DAPFitsUtil.read(str(ifile), checksum=checksum)

        # Confirm that the internal method is the same as the method
        # that was used in writing the file
        if self.hdu['PRIMARY'].header['SCKEY'] != self.method['key']:
            if strict:
                raise ValueError('Keywords in header do not match specified method keywords!')
            elif not self.quiet:
                warnings.warn('Keywords in header do not match specified method keywords!')
        # TODO: "strict" should also check other aspects of the file to
        # make sure that the details of the method are also the same,
        # not just the keyword

        # Attempt to read the input guess velocity and sigma
        try:
            self.guess_vel = self.hdu['PRIMARY'].header['SCINPVEL']
            self.guess_sig = self.hdu['PRIMARY'].header['SCINPSIG']
        except:
            if not self.quiet:
                warnings.warn('Unable to read input guess kinematics from file header!')
            self.guess_vel = None
            self.guess_sig = None

        # Attempt to read the modeling parameters
        if self.method['fitpar'] is not None and callable(self.method['fitpar'].fromheader):
            self.method['fitpar'].fromheader(self.hdu['PRIMARY'].header)

        self.nmodels = self.hdu['FLUX'].shape[0]
        self.missing_models = self._get_missing_models(debug=debug)

    def copy_to_array(self, ext='FLUX', waverange=None, include_missing=False):
        r"""
        Wrapper for :func:`mangadap.util.fitsutil.DAPFitsUtil.copy_to_array`
        specific for :class:`StellarContinuumModel`.

        Return a copy of the selected data array. The array size is
        always :math:`N_{\rm models} \times N_{\rm wavelength}`; i.e.,
        the data is always flattened to two dimensions and the unique
        spectra are selected.

        Args:
            ext (:obj:`str`, optional):
                Name of the extension from which to draw the data.
                Must be allowed for the current :attr:`mode`; see
                :attr:`data_arrays`.
            waverange (array-like, optional):
                Two-element array with the first and last wavelength
                to include in the computation. If None, use the full
                wavelength range.
            include_missing (:obj:`bool`, optional):
                Create an array with a size that accommodates the
                missing models.

        Returns:
            `numpy.ndarray`_: A 2D array with a copy of the data from
            the selected extension.
        """
        return DAPFitsUtil.copy_to_array(self.hdu, ext=ext, allowed_ext=self.spectral_arrays,
                                         waverange=waverange,
                                    missing_bins=self.missing_models if include_missing else None,
                                         nbins=self.nbins,
                                    unique_bins=DAPFitsUtil.unique_bins(self.hdu['BINID'].data))

    def copy_to_masked_array(self, ext='FLUX', flag=None, waverange=None, include_missing=False):
        """
        Wrapper for
        :func:`mangadap.util.fitsutil.DAPFitsUtil.copy_to_masked_array`
        specific for :class:`StellarContinuumModel`.

        Return a copy of the selected data array as a masked array.
        This is functionally identical to :func:`copy_to_array`,
        except the output format is a `numpy.ma.MaskedArray`_. The
        pixels that are considered to be masked can be specified
        using `flag`.
        
        Args:
            ext (:obj:`str`, optional):
                Name of the extension from which to draw the data.
                Must be allowed for the current :attr:`mode`; see
                :attr:`data_arrays`.
            flag (:obj:`str`, :obj:`list`, optional):
                (List of) Flag names that are considered when
                deciding if a pixel should be masked. The names
                *must* be a valid bit name as defined by
                :attr:`bitmask`. If not provided, *ANY* non-zero mask
                bit is omitted.
            waverange (array-like, optional):
                Two-element array with the first and last wavelength
                to include in the computation. If None, use the full
                wavelength range.
            include_missing (:obj:`bool`, optional):
                Create an array with a size that accommodates the
                missing models.

        Returns:
            `numpy.ma.MaskedArray`_: A 2D array with a copy of the
            data from the selected extension.
        """
        return DAPFitsUtil.copy_to_masked_array(self.hdu, ext=ext, mask_ext='MASK', flag=flag,
                                                bitmask=self.bitmask,
                                                allowed_ext=self.spectral_arrays,
                                                waverange=waverange,
                                    missing_bins=self.missing_models if include_missing else None,
                                                nbins=self.nmodels,
                                        unique_bins=DAPFitsUtil.unique_bins(self.hdu['BINID'].data))

    def construct_models(self, replacement_templates=None, redshift_only=False,
                         deredshift=False, corrected_dispersion=False):
        """
        Reconstruct the best fitting models.

        This is largely a wrapper for a :func:`construct_models`
        method provided by the :attr:`method` fitting class. As such,
        this method will raise an exception if the fitting class has
        no such method.

        Args:
            replacement_templates (:class:`~mangadap.proc.templatelibary.TemplateLibrary`, optional):
                Instead of using the template library used to fit the
                data, use this library instead. The number of spectra
                in the replacement library **must** match the
                original number of templates. This is only really
                useful when replacing the original set of templates
                with ones at a different spectral resolution.
            redshift_only (:obj:`bool`, optional):
                When constructing the models, only redshift the
                spectra, effectively providing the best-fitting model
                with a velocity dispersion of 0 km/s.
            deredshift (:obj:`bool`, optional):
                Return the best-fitting models in the rest frame.
            corrected_dispersion (:obj:`bool`, optional):
                Return the best-fitting models with a LOSVD that has
                been corrected for any resolution difference between
                the fitted templates and the binned spectra.

        Returns:
            `numpy.ndarray`_: The best-fitting models.

        Raises:
            ValueError: 
                Raised if there is no defined fitting class.
            AttributeError:
                Raised if the fitting class has no method called
                :func:`construct_models`.
            TypeError:
                Raised if the provided ``replacement_templates`` are
                not an instance of
                :class:`mangadap.proc.templatelibary.TemplateLibrary`.
        """
        if self.method['fitclass'] is None:
            raise ValueError('No class object available for constructing the model!')
        if not callable(self.method['fitclass'].construct_models):
            raise AttributeError('Provided fit class object has no callable ' \
                                 '\'construct_models\' attribute!')
        if replacement_templates is not None \
                and not isinstance(replacement_templates, TemplateLibrary):
            raise TypeError('Provided template replacements must have type TemplateLibrary.')

        # Only the selected models are constructed, others are masked
        select = numpy.invert(self.bitmask.flagged(self.hdu['PAR'].data['MASK'],
                                            flag=['NO_FIT', 'FIT_FAILED', 'INSUFFICIENT_DATA']))
        templates = self.method['fitpar']['template_library'] if replacement_templates is None \
                                            else replacement_templates

        # The only reason it needs the flux data is to get the shape, so
        # it doesn't matter that I'm passing the model flux
        return self.method['fitclass'].construct_models(templates['WAVE'].data,
                                                        templates['FLUX'].data,
                                                        self.hdu['WAVE'].data,
                                                        self.hdu['FLUX'].data.shape,
                                                        self.hdu['PAR'].data, select=select,
                                                        redshift_only=redshift_only,
                                                        deredshift=deredshift,
                                                        corrected_dispersion=corrected_dispersion,
                                                        dvtol=1e-9)

    # TODO: Get rid of dispaxis?
    # TODO: Make this a general routine.
    @staticmethod
    def reset_continuum_mask_window(continuum, dispaxis=1, quiet=False):
        """
        Reset the mask of the stellar continuum to a continuous window
        from the minimum to maximum valid wavelength.

        .. todo::
            - Allow continuum to be n-dimensional and use
              `numpy.apply_along_axis`.

        Args:   
            continuum (`numpy.ma.MaskedArray`_):
                Stellar continuum models.
            dispaxis (:obj:`int`, optional):
                Wavelength axis in provided ``continuum`` object.
            quiet (:obj:`bool`, optional):
                Suppress terminal output.

        Returns:
            `numpy.ma.MaskedArray`_: Continuum object but unmasked
            over the full continuous window between the minimum and
            maximum valid wavelength.
        """
        spatial_shape = DAPFitsUtil.get_spatial_shape(continuum.shape, dispaxis)
        if len(spatial_shape) != 1:
            raise ValueError('Input array should be two-dimensional!')

        # No pixels are masked!
        if numpy.sum(numpy.ma.getmaskarray(continuum)) == 0:
            return continuum

        _continuum = continuum.copy() if dispaxis == 1 else continuum.copy().T

        nspec,npix = _continuum.shape
        pix = numpy.ma.MaskedArray(numpy.array([ numpy.arange(npix) ]*nspec),
                                   mask=numpy.ma.getmaskarray(_continuum).copy())

        min_good_pix = numpy.ma.amin(pix, axis=dispaxis)
        max_good_pix = numpy.ma.amax(pix, axis=dispaxis)
        for c,s,e in zip(_continuum, min_good_pix,max_good_pix+1):
            if isinstance(s, numpy.ma.core.MaskedConstant) \
                    or isinstance(e, numpy.ma.core.MaskedConstant):
                continue
            numpy.ma.getmaskarray(c)[s:e] = False

        return _continuum if dispaxis == 1 else _continuum.T


    def unmasked_continuum_model(self, replacement_templates=None, redshift_only=False,
                                 deredshift=False, corrected_dispersion=False):
        """
        Return the stellar continuum unmasked over the full continuous
        fitting region.

        Models are reconstructed based on the model parameters if new
        template fluxes are provided or if the returned models should
        not include the LOSVD convolution (redshift_only=True). As
        such, this is largely a wrapper for :func:`construct_models`
        and :func:`reset_continuum_mask_window`.

        Args:
            replacement_templates (:class:`mangadap.proc.templatelibary.TemplateLibrary`, optional):
                Instead of using the template library used to fit the
                data, use this library instead. The number of spectra
                in the replacement library **must** match the
                original number of templates. This is only really
                useful when replacing the original set of templates
                with ones at a different spectral resolution.
            redshift_only (:obj:`bool`, optional):
                When constructing the models, only redshift the
                spectra, effectively providing the best-fitting model
                with a velocity dispersion of 0 km/s.
            deredshift (:obj:`bool`, optional):
                Return the best-fitting models in the rest frame.
            corrected_dispersion (:obj:`bool`, optional):
                Return the best-fitting models with a LOSVD that has
                been corrected for any resolution difference between
                the fitted templates and the binned spectra.

        Returns:
            `numpy.ndarray`_: The best-fitting models.
        """
        # Get the models for the binned spectra
        reconstruct = replacement_templates is not None or redshift_only or corrected_dispersion
        continuum = self.construct_models(replacement_templates=replacement_templates,
                                          redshift_only=redshift_only, deredshift=deredshift,
                                          corrected_dispersion=corrected_dispersion) \
                        if reconstruct \
                        else self.copy_to_masked_array(flag=self.all_spectrum_flags())
        return StellarContinuumModel.reset_continuum_mask_window(continuum, quiet=self.quiet)

    def fill_to_match(self, binid, replacement_templates=None, redshift_only=False,
                      deredshift=False, corrected_dispersion=False, missing=None):
        """
        Get the stellar-continuum model from this objects that matches
        the input bin ID matrix.

        The output is a 2D matrix ordered by the bin ID; any skipped
        index numbers in the maximum of the union of the unique
        numbers in the ``binid`` and ``missing`` input are masked.

        All this does is find the stellar continuum from spaxels in
        this model and match them to the input spaxels. Depending on
        how differently the data sets have been binned, this says
        nothing about whether or not the returned models are a good
        fit to the data!

        Args:
            binid (`numpy.ndarray`_):
                A map of the bin IDs with shape that matches the
                spatial shape of the parent datacube; see
                :attr:`spatial_shape`. This provides the mapping
                between the current stellar continuum models and the
                output data array.
            replacement_templates (:class:`mangadap.proc.templatelibary.TemplateLibrary`, optional):
                Instead of using the template library used to fit the
                data, use this library instead. The number of spectra
                in the replacement library **must** match the
                original number of templates. This is only really
                useful when replacing the original set of templates
                with ones at a different spectral resolution.
            redshift_only (:obj:`bool`, optional):
                When constructing the models, only redshift the
                spectra, effectively providing the best-fitting model
                with a velocity dispersion of 0 km/s.
            deredshift (:obj:`bool`, optional):
                Return the best-fitting models in the rest frame.
            corrected_dispersion (:obj:`bool`, optional):
                Return the best-fitting models with a LOSVD that has
                been corrected for any resolution difference between
                the fitted templates and the binned spectra.
            missing (array-like, optional):
                Include empty spectra for binids with these numbers.
                Ignored if None or an empty array.

        Returns:
            `numpy.ndarray`_: The best-fitting models matched to the
            input bin ID array.

        Raises:
            ValueError:
                Raised if the shape of ``binid`` does not match
                :attr:`spatial_shape`.
        """
        # The input bin id array must have the same shape as self
        if binid.shape != self.spatial_shape:
            raise ValueError('Input bin ID matrix has incorrect shape.')

        # Construct the best-fitting models
        best_fit_continuum = self.unmasked_continuum_model(
                                    replacement_templates=replacement_templates,
                                    redshift_only=redshift_only, deredshift=deredshift,
                                    corrected_dispersion=corrected_dispersion)

        # Get the number of output continuua
        nbins = numpy.amax(binid).astype(int)+1 if missing is None or len(missing) == 0 else \
                    numpy.amax( numpy.append(binid.ravel(), missing) ).astype(int)+1

        # Map the BINID to the spectrum index, assuming bins are sorted
        u, indx, reconstruct = numpy.unique(self['BINID'].data.ravel(), return_index=True,
                                            return_inverse=True)
#        u_bin_indx = numpy.arange(len(u))-1
        u_bin_indx = numpy.arange(u.size)
        if u[0] == -1:
            u_bin_indx -= 1
        _bin_indx = u_bin_indx[reconstruct].reshape(self.spatial_shape)

        # Fill in bins with no models with masked zeros
        continuum = numpy.ma.zeros((nbins,best_fit_continuum.shape[1]), dtype=float)
        continuum[:,:] = numpy.ma.masked
        for i,j in zip(binid.ravel(), _bin_indx.ravel()):
            if i < 0 or j < 0:
                continue
            continuum[i,:] = best_fit_continuum[j,:]

        return continuum

    def matched_kinematics(self, binid, redshift=None, dispersion=100.0, constant=False, cz=False,
                           corrected=False, min_dispersion=None, nearest=False, missing=None):
        r"""
        Get the stellar kinematics from this objects that matches the
        input bin ID matrix.

        This method has a similar intent to :func:`fill_to_match`,
        but instead of matching the model spectra, its intent is to
        match the stellar kinematics.

        For spectra the were not fit but have a bin ID (see
        ``binid``), the method uses the median (unmasked) kinematic
        measurement if no default values are provided (see
        ``redshift`` and ``dispersion``).

        Args:
            binid (`numpy.ndarray`_):
                2D array with the bin ID associate with each spaxel
                in the datacube. Shape must be the same as
                :attr:`spatial_shape`.
            redshift (:obj:`float`, optional):
                The default redshift to use for spectra without a
                stellar-continuum fit. Default is to use the median
                of the unmasked measurements
            dispersion (:obj:`float`, optional):
                The default velocity dispersion to use for spectra
                without a stellar-continuum fit. Default is 100 km/s.
            constant (:obj:`bool`, optional):
                Force the function to return a constant redshift and
                dispersion for each spectrum, regardless of any
                fitted kinematics.
            cz (:obj:`bool`, optional):
                Return the redshift as cz in km/s, as opposed to
                unitless z.
            corrected (:obj:`bool`, optional):
                Return the velocity dispersions corrected for any
                resolution difference between the templates and the
                galaxy data. If False, return the uncorrected data.
            min_dispersion (:obj:`float`, optional):
                Impose a minimum dispersion.
            nearest (:obj:`bool`, optional):
                Instead of the median of the results for the spectra
                that were not fit, use the value from the nearest
                bin.
            missing (array-like, optional):
                Any bin ID numbers missing from the input ``binid``
                image needed for constructing the output matched
                data.

        Returns:
            :obj:`tuple`: Returns `numpy.ndarray`_ objects with a
            redshift (or cz) and dispersion for each binned spectrum.
            Shape is :math:`(N_{\rm bin},)`.

        Raises:
            TypeError:
                Raised if the input redshift or dispersion values are
                not single numbers.
            ValueError:
                Raised if the shape of ``binid`` does not match
                :attr:`spatial_shape`.
        """
        # Check input
        if redshift is not None and not isinstance(redshift, (float,int)):
            raise TypeError('Redshift must be a single number or None.')
        if dispersion is not None and not isinstance(dispersion, (float,int)):
            raise TypeError('Dispersion must be a single number or None.')
        if binid.shape != self.spatial_shape:
            raise ValueError('Input bin ID map must match the spatial shape of the DRP cube.')

        # Get the number of output kinematics
        nbins = numpy.amax(binid).astype(int)+1 if missing is None or len(missing) == 0 else \
                    numpy.amax( numpy.append(binid.ravel(), missing) ).astype(int)+1

        # Mask bad values
        mask = self.bitmask.flagged(self.hdu['PAR'].data['MASK'].copy(),
                                    [ 'NO_FIT', 'FIT_FAILED', 'INSUFFICIENT_DATA', 'NEAR_BOUND' ])

        if numpy.all(mask):
            # Everything is masked so use input guesses
            warnings.warn('All stellar continuum fits have been masked!  Using input guesses.')
            str_z = numpy.ma.MaskedArray(self.method['fitpar']['guess_redshift'].copy())
            str_d = numpy.ma.MaskedArray(self.method['fitpar']['guess_dispersion'].copy())
            # Select one per *stellar-continuum model*
            str_z = str_z[self['PAR'].data['BINID_INDEX']]
            str_d = str_d[self['PAR'].data['BINID_INDEX']]
        else:
            # Pull out the data
            str_z = numpy.ma.MaskedArray(self.hdu['PAR'].data['KIN'][:,0].copy(), mask=mask) \
                            / astropy.constants.c.to('km/s').value
            str_d = numpy.ma.MaskedArray(self.hdu['PAR'].data['KIN'][:,1].copy(), mask=mask)

            # Apply the sigma corrections if requested.  Values with the
            # correction larger than the measured value will be masked!
            if corrected:
                sigma_corr = numpy.ma.MaskedArray(self.hdu['PAR'].data['SIGMACORR_EMP'], mask=mask)
                str_d = numpy.ma.sqrt( numpy.square(str_d) - numpy.square(sigma_corr) )

        # Set the default values to use when necessary
        _redshift = numpy.ma.median(str_z) if redshift is None else redshift
        _dispersion = numpy.ma.median(str_d) if dispersion is None else dispersion
        if _dispersion is numpy.ma.masked:
            _dispersion = numpy.median(self.method['fitpar']['guess_dispersion'])
        if min_dispersion is not None and _dispersion < min_dispersion:
            _dispersion = min_dispersion

        # Just return the single value
        if constant:
            str_z = numpy.full(nbins, _redshift * astropy.constants.c.to('km/s').value
                                        if cz else _redshift, dtype=float)
            str_d = numpy.full(nbins, _dispersion, dtype=float)
            return str_z, str_d

        # Fill masked values with the nearest datum
        if nearest:
            # Fill masked values with the nearest datum
            replace = numpy.ma.getmaskarray(str_z) | numpy.ma.getmaskarray(str_d)

            if numpy.all(replace):
                # All are bad so replace with _redshift and _dispersion
                str_z = numpy.full(str_z.shape, _redshift, dtype=float)
                str_d = numpy.full(str_d.shape, _dispersion, dtype=float)
            elif numpy.any(replace):
                best_fit_kinematics = numpy.ma.append([str_z], [str_d], axis=0).T
                valid_bins = numpy.unique(self['BINID'].data)
                if valid_bins[0] == -1:
                    valid_bins = valid_bins[1:]
                coo = self.binned_spectra['BINS'].data['SKY_COO'][valid_bins,:]
                kinematics = replace_with_data_from_nearest_coo(coo, best_fit_kinematics, replace)
                str_z = kinematics[:,0]
                str_d = kinematics[:,1]
        else:
            # Fill any masked values with the single estimate
            str_z = str_z.filled(_redshift)
            str_d = str_d.filled(_dispersion)

        # Map the BINID to the spectrum index, assuming bins are sorted
        # and that the BINID map has -1 BINID values
        u, indx, reconstruct = numpy.unique(self['BINID'].data.ravel(), return_index=True,
                                            return_inverse=True)
#        u_bin_indx = numpy.arange(len(u))-1
        u_bin_indx = numpy.arange(u.size)
        if u[0] == -1:
            u_bin_indx -= 1
        _bin_indx = u_bin_indx[reconstruct].reshape(self.spatial_shape)

        # Match the kinematics to the output bin ID map
        _str_z = numpy.ma.masked_all(nbins, dtype=float)
        _str_d = numpy.ma.masked_all(nbins, dtype=float)
        for i,j in zip(binid.ravel(), _bin_indx.ravel()):
            if i < 0 or j < 0:
                continue
            _str_z[i] = str_z[j]
            _str_d[i] = str_d[j]

        str_z = _str_z.filled(_redshift)
        str_d = _str_d.filled(_dispersion)
        # Convert to cz velocities (km/s)
        if cz:
            str_z *= astropy.constants.c.to('km/s').value
        return str_z, str_d

# TODO: Function out of date!  Keep it around for now though...
#    def matched_template_flags(self, binned_spectra):
#        """
#        Return the templates used during the fit to each spectrum,
#        matched to the spectra in the binned_spectra object.  For
#        spectra with no models, the flags are all set to true.
#        """
#        if binned_spectra is self.binned_spectra:
#            usetpl = self.hdu['PAR'].data['TPLWGT'] > 0
#
#            # Number of models matches the numbers of bins
#            if binned_spectra.nbins == self.nmodels:
##                print('returning usetpl')
#                return usetpl
#    
#            # Fill in bins with no models with masked zeros
##            print('Fill in bins with no models with masked zeros')
#            ntpl = self.method['fitpar']['template_library'].ntpl
#            _usetpl = numpy.ones((binned_spectra.nbins,ntpl), dtype=bool)
#            for i,j in enumerate(self.hdu['PAR'].data['BINID_INDEX']):
#                _usetpl[j,:] = usetpl[i,:]
#            return _usetpl
#
#        raise NotImplementedError('Can only match to internal binned_spectra.')




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Container class that defines a bandpass filter.

Class usage examples
--------------------

To define an bandpass filter::

    from mangadap.par.bandpassfilter import BandPassFilterPar
    p = BandPassFilterPar(index=44, name='Ha', blueside=[6483.0,6513.0],
                          redside=[6623.0,6653.0], restwave=6564.632,
                          primary=[6557.6,6571.6]) 

However, this class is mostly to provide a base class used by
:class:`mangadap.par.emissionmomentsdb.EmissionMomentsDB`,
:class:`mangadap.par.absorptionindexdb.AbsorptionIndexDB`, and
:class:`mangadap.par.bandheadindexdb.BandheadIndexDB`; it is not
really meant to be used as given above.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""

import warnings

from IPython import embed

import numpy

from ..par.parset import KeywordParSet
from ..util.sampling import grid_borders

from matplotlib import pyplot

# Add strict versioning
# from distutils.version import StrictVersion

class BandPassFilterPar(KeywordParSet):
    r"""
    Parameter object that defines a set of band-pass filters.  This is a
    base class for similar objects used calculating fluxes and flux
    moments of emission lines
    (:class:`mangadap.par.emissionmomentsdb.EmissionMomentsDB`),
    absorption line indices
    (:class:`mangadap.par.absorptionindexdb.AbsorptionIndexDB`), and
    spectral continuum bandheads
    (:class:`mangadap.par.bandheadindexdb.BandheadIndexDB`) in the DAP.

    All wavelengths are expected to be **in vacuum**, to match the
    expected application of this filter to SDSS MaNGA spectra.

    See :class:`mangadap.par.parset.ParSet` for attributes and raised
    exceptions.

    The defined parameters are:

    .. include:: ../tables/bandpassfilterpar.rst
    """
    def __init__(self, index=None, name=None, blueside=None, redside=None, restwave=None,
                 primary=None, units=None, component=None, integrand=None, order=None):
        
        in_fl = [ int, float ]
        ar_like = [ numpy.ndarray, list ]
        n = None if name is None else name.strip()
        unit_opts = [ 'ang', 'mag' ]
        integ_opts = [ 'flambda', 'fnu' ]
        order_opts = [ 'b_r', 'r_b' ]
        
        pars  =    [ 'index', 'name', 'restwave', 'primary', 'blueside', 'redside',   'units',
                     'component', 'integrand',    'order' ]
        values  =  [   index,      n,   restwave,   primary,   blueside,   redside,     units,
                       component,   integrand,      order ]
        options  = [    None,   None,       None,      None,       None,      None, unit_opts,
                            None,  integ_opts, order_opts ]
        dtypes  =  [     int,    str,      in_fl,   ar_like,    ar_like,   ar_like,       str,
                            bool,         str,        str ]

        descr = ['An index used to refer to the bandpass filter',
                 'A name for the bandpass filter',
                 'The rest wavelength of the line in the primary passband in angstroms **in ' \
                    'vacuum**.  This is used to convert the first and second moments to ' \
                    'velocity (km/s) for emission lines.',
                 'A two-element vector with the starting and ending wavelength (angstroms in ' \
                    '**vacuum**) for a passband to the blue of the primary spectral feature.',
                 'A two-element vector with the starting and ending wavelength (angstroms in ' \
                    '**vacuum**) for a passband to the red of the primary spectral feature.',
                 'A two-element vector with the starting and ending wavelength (angstroms in ' \
                    '**vacuum**) for the primary passband surrounding the spectral feature of ' \
                    'interest.  This is used by ' \
                    ':class:`mangadap.par.emissionmomentsdb.EmissionMomentsDB` and ' \
                    ':class:`mangadap.par.absorptionindexdb.AbsorptionIndexDB`, but it is ' \
                    'irrelevant for :class:`mangadap.par.bandheadindexdb.BandheadIndexDB`.',
                 'Define the unit for the spectral index as either angstroms (``ang``) or ' \
                    'magnitudes (``mag``).  Currently only used by ' \
                    ':class:`mangadap.par.absorptionindexdb.AbsorptionIndexDB`.',
                 'Flag that the bandpass definition is a component of a larger set of bandpass ' \
                    'filters.  This is currently only used by ' \
                    ':class:`mangadap.par.absorptionindexdb.AbsorptionIndexDB` to combine index ' \
                    'measurements into a single index.  If True, all components with the same ' \
                    '*name* are summed to form the composite index.',
                 r'Currently only used by :class:`mangadap.par.bandheadindexdb.BandheadIndexDB`.' \
                    r'  Define the integrand over the passband used to construct and index as ' \
                    r'either flux per unit frequency (``fnu``), :math:`F_\nu`, or flux per unit ' \
                    r'wavelength (``flambda``), :math:`F_\lambda`.',
                 'Currently only used by :class:`mangadap.par.bandheadindexdb.BandheadIndexDB`.' \
                    '  Define the order to use when constructing the index.  The options are ' \
                    'either a ratio of red-to-blue or blue-to-red, which are respectively ' \
                    'selected using ``r_b`` or ``b_r``.']

        super(BandPassFilterPar, self).__init__(pars, values=values, options=options,
                                                dtypes=dtypes, descr=descr)
        self._check()

    def __repr__(self):
        """Return a string representation of the bandpass filter."""
        summary = 'index     : {0}'.format(self['index'])
        summary += '\nname      : {0}'.format(self['name'])
        if self['restwave'] is not None:
            summary += '\nrestwave  : {0}'.format(self['restwave'])
        if self['primary'] is not None:
            summary += '\nprimary   : {0}'.format(self['primary'])
        summary += '\nblueside  : {0}'.format(self['blueside'])
        summary += '\nredside   : {0}'.format(self['redside'])
        if self['units'] is not None:
            summary += '\nunits     : {0}'.format(self['units'])
        if self['component'] is not None:
            summary += '\ncomponent : {0}'.format(self['component'])
        if self['integrand'] is not None:
            summary += '\nintegrand : {0}'.format(self['integrand'])
        if self['order'] is not None:
            summary += '\norder     : {0}'.format(self['order'])
        return summary

    def _check(self):
        """
        Check the parameter list:
            
            - Make sure arrays only have two elements.

        .. todo::
            - Add check to __setitem__()?

        Raises:
            ValueError: Raised if one of the conditions above are not
                met.
        """
        if self.data['primary'] is not None and len(self.data['primary']) != 2:
            raise ValueError('Primary passband must have two and only two elements.')
        if self.data['blueside'] is not None and len(self.data['blueside']) != 2:
            raise ValueError('Blue sideband must have two and only two elements.')
        if self.data['redside'] is not None and len(self.data['redside']) != 2:
            raise ValueError('Red sideband must have two and only two elements.')


def passband_median(x, y, passband=None):
    """
    Determine the median of the y values within the passband
    """

    # Ensure that y is a numpy array and compress it if it's a masked
    # array
    _x = x if isinstance(x, numpy.ma.MaskedArray) else numpy.ma.MaskedArray(x)
    _y = y if isinstance(y, numpy.ma.MaskedArray) else numpy.ma.MaskedArray(y)
    if len(_x.shape) != 1 or len(_y.shape) != 1:
        raise ValueError('Can only provide one-dimensional vectors.')
    mask = numpy.ma.getmaskarray(_x) | numpy.ma.getmaskarray(_y)
    _y[mask] = numpy.ma.masked
    _y = _y.compressed()
    if _y.size == 0:
        warnings.warn('No elements remain in the passband.')
        return 0.0

    if passband is None:
        return numpy.ma.median(_y)

    _x[mask] = numpy.ma.masked
    _x = _x.compressed()

    indx = numpy.array([numpy.where(numpy.logical_and(_x > p[0], _x < p[1]))[0]
                                for p in passband], dtype=object)
    nonzero = numpy.array([ len(ii) > 0 for ii in indx ])
    if not numpy.all(nonzero):
        warnings.warn('Returning empty passbands with median values of 0!')
    return numpy.array([ 0.0 if len(ii) == 0 else numpy.ma.median(_y[ii]) for ii in indx ])


def pixel_fraction_in_passband(x, passband, dx=None):
    r"""
    Determine the width of each x interval to include in one or more
    passband integrals.

    Args:
        x (array-like):
            1D array with pixel borders.
        passband (array-like):
            1D or 2D array with the passbands to integrate over. If
            more than one passband is provided, the shape of the
            input should be :math:`(N_{\rm passband},2)`.
        dx (array-like, optional):
            1D array with the size of each pixel. If not provided,
            calculated from `x`. Can be provided to speed
            calculations for multiple calls.

    Returns:
        `numpy.ndarray`_: The array with the fraction of each pixel
        within each passband. The output shape is :math:`(N_{\rm
        pixel},N_{\rm passband})`, where :math:`N_{\rm pixel}` is the
        1 less than the number of pixel borders.
    
    """
    _x = numpy.atleast_1d(x)
    if _x.ndim != 1:
        raise ValueError('Pixel coordinates must be a 1D vector.')
    _dx = numpy.diff(_x) if dx is None else numpy.atleast_1d(dx)
    if _dx.ndim != 1:
        raise ValueError('Pixel sizes must be a 1D vector.')
    if _x.size != _dx.size+1:
        raise ValueError('Input coordinates must be the pixel borders.')
    _passband = numpy.atleast_1d(passband)
    if _passband.ndim == 1:
        return ((_x[1:]-_passband[0])/_dx).clip(0,1)+((_passband[1]-_x[:-1])/_dx).clip(0,1)-1.0
    
    return ((x[1:,None]-passband[None,:,0])/dx[:,None]).clip(0,1) \
                + ((passband[None,:,1]-x[:-1,None])/dx[:,None]).clip(0,1) - 1.0


def passband_integral(x, y, passband=None, borders=False, log=False, base=10.0, quad=False):
    r"""
    Determine the integral of a discrete set of data over a passband
    accounting for fractional pixels.

    The integral is done by a simple sum:
    
    .. math::
        S(y) = \sum_i \delta_i y_i {\rm d}x_i
        
    where :math:`\delta_i` is the fraction of pixel :math:`i` within the
    integration limits as determined by
    :func:`pixel_fraction_in_passband`.  Given the errors in :math:`y`
    (:math:`\epsilon_i`), the propagated error in :math:`S` is

    .. math::

        [\epsilon S(y)]^2 = \sum_i (\delta_i \epsilon_i {\rm d}x_i)^2, 

    which can be calculated using the ``quad`` keyword. For example,
    given values to integrate, ``y``, and errors, ``e``, calculate
    the passband integral and its error as follows::

        integral = passband_integral(x, y)
        integral_error = passband_integral(x, e, quad=True)

    .. warning::

        Although the equation for the formal error propagation is
        correct, a comparison of the formal errors to errors from a
        Monte Carlo simulation show that the former are pretty poor
        estimates of the latter.

    Args:
        x (array-like):
            The 1D array of x coordinates for the integration. If
            borders=True, this array provides the borders of the
            interval over which y is measured. These can be
            non-uniform. In this case, the number of x values must be
            :math:`N+1` for :math:`N` 1D ``y`` values (see below).
            Otherwise, the `x` values are expected to be uniformly
            sampled either linearly or geometrically. In this case,
            ``x`` can be either a two element vector giving the
            (geometric) centers of the first and last sample interval
            or a vector with the :math:`N` samples. In either case,
            :func:`mangadap.util.sampling.grid_borders` is used to
            determine the borders of the sample intervals.
        y (array-like):
            The 1D or 2D array of measured values to be integrated.
            The first axis must always provide the values to be
            integrated. If those values are specific to the passband,
            ``y`` should be 2D with shape :math:`(N_y,N_{\rm
            passband})`.
        passband (array-like, optional):
            The 1D or 2D array with the set of passbands (in units of
            x) to integrate. The length of the last dimension must
            always be 2, providing the starting and ending x
            coordinate. If 2D, the array must have shape
            :math:`(N_{\rm passband},2)`. If not provided, the
            integral is just performed over the full :math:`y`
            vector. If ``y`` is 2D, the last axis of ``y`` and the
            first axis here must match.
        borders (:obj:`bool`, optional):
            The :math:`x` values provided are the border coordinates
            over which the :math:`y` values have been determined.
        log (:obj:`bool`, optional):
            The coordinates are logarithmically sampled. Ignored if
            ``borders`` is True.
        base (:obj:`float`, optional):
            The base of the logarithm used in the geometric sampling
            if ``log`` is True.
        quad (:obj:`bool`, optional):
            Perform the quadrature sum instead of the direct sum.  This
            is used for error calculations.
        
    Returns:
        :obj:`float`, `numpy.ndarray`_: The passband integral for
        each passband.

    Raises:
        ValueError:
            Raised if ``x`` is not a 1D array, if the shapes of ``x``
            and ``y`` are not appropriately matched, if ``passbands``
            has more than 2 dimensions, if the last axis of
            ``passbands`` is not 2 elements long, or if the shapes of
            ``passbands`` and ``y`` are not appropriately matched.
    """
    # Ensure that y is a numpy array, fill any masked values with 0s,
    # and check its shape
    _y = y.filled(0.0) if isinstance(y, numpy.ma.MaskedArray) else numpy.atleast_1d(y) 

    if passband is None and _y.ndim > 1:
        raise ValueError('If no passbands provided, must provide single vector to integrate.')

    ny = _y.shape[0]

    # Check the input coordinates and calculate the borders
    _x = numpy.atleast_1d(x)
    if len(_x.shape) != 1:
        raise ValueError('Coordinates must be a 1D vector.')
    if borders and len(_x) != ny+1:
        raise ValueError('Must provide N+1 borders for N samples.')
    if not borders:
        _x, dx = grid_borders(numpy.array([x[0],x[-1]]), ny, log=log, base=base)

    if passband is None:
        # No passband, so just get the full integral; y is checked to
        # be 1D above
        return numpy.sqrt(numpy.dot(numpy.square(_y), numpy.square(numpy.diff(_x)))) if quad \
                    else numpy.dot(_y, numpy.diff(_x))

    # Check the input passbands
    _passband = numpy.atleast_2d(passband)
    if _passband.ndim > 2:
        raise ValueError('Can only handle passband arrays of 1 or 2 dimensions.')
    if _passband.shape[-1] != 2:
        raise ValueError('Passband(s) must have shape N_bands x 2.')
    
    # Reshape y if necessary
    _y = numpy.atleast_2d(_y)
    if _y.shape[0] == 1:
        _y = numpy.tile(_y[0], (_passband.shape[0],1)).T

    # Check its shape
    if _y.shape[1] != _passband.shape[0]:
        raise ValueError('For 2D array input, second axis must match the number of passbands.')
        
    # Get the (fractional) interval of each pixel in each passband
    dx = numpy.diff(_x)
    inband_dx = pixel_fraction_in_passband(_x, _passband, dx)*dx[:,None]

    # Sum of y*dx, or sqrt(sum((y*dx)^2))
    integral = numpy.sqrt(numpy.sum(numpy.square(_y*inband_dx), axis=0)) if quad \
                    else numpy.sum(_y*inband_dx, axis=0)

    # Reshape the returned array if only one passband provided
    return integral if numpy.atleast_1d(passband).ndim == 2 else integral[0]


def passband_integrated_width(x, y, passband=None, borders=False, log=False, base=10.0):
    """
    Determine the integrated width of the passband, accounting for masked pixels.
    """
    unity = y.copy()
    unity[numpy.invert(numpy.ma.getmaskarray(y))] = 1.0
    return passband_integral(x, unity, passband=passband, borders=borders, log=log, base=base)


def passband_integrated_mean(x, y, err=None, passband=None, borders=False, log=False, base=10.0):
    """
    Determine the integrated mean over a (set of) passband(s).  Nominal
    errors are returned if err is provided.
    """
    # Get the passband interval, accounting for masked pixels
    interval = passband_integrated_width(x, y, passband=passband, borders=borders, log=log,
                                         base=base)
    interval = numpy.ma.MaskedArray(interval, mask=numpy.invert(numpy.absolute(interval) > 0))

    # Get the integral of y over the passband
    integral = numpy.ma.MaskedArray(passband_integral(x, y, passband=passband, borders=borders,
                                                      log=log, base=base))
    if err is None:
        # No error
        return numpy.ma.divide(integral, interval), None

    if len(y) != len(err):
        raise ValueError('Value and error have different sizes.')

    # Return the integrated mean and error
    integral_err = numpy.ma.MaskedArray(passband_integral(x, err, passband=passband,
                                                          borders=borders, log=log, base=base,
                                                          quad=True))
    return numpy.ma.divide(integral, interval), numpy.ma.divide(integral_err, interval)


def passband_weighted_mean(x, y, z, passband=None, borders=False, yerr=None, zerr=None, log=False,
                           base=10.0):
    r"""
    Determine the y-weighted mean of z over the passband; i.e.,

    .. math::

        \mu = \frac{\int_{x_1}^{x_2} y z dx}{\int_{x_1}_{x_2} y dx},
        
    where the integrals are determined by the discrete sum; i.e.,
    :math:`\mu = S(y z)/S(y)` where :math:`S(y)` is provided by
    :func:`passband_integral`.

    Because the error in the numerator and denominator of :math:`\mu`
    are correlated, they are calculated here by explicit error
    propagation:

    .. math::

        \epsilon_\mu^2 = \frac{1}{S(y)^2} (S((z-\mu)^2 \epsilon_y^2) +
        S(y^2\epsilon_z^2)).

    .. todo::
        -doc the args

    Returns:
        `numpy.ma.MaskedArray`_: Two masked arrays with the
        passband-weighted mean and its error. The error is returned
        as `None` if no errors are provided.
    """
    # Get the passband interval, accounting for masked pixels
    weighted_integral = passband_integral(x, z*y, passband=passband, borders=borders, log=log,
                                          base=base)
    integral = passband_integral(x, y, passband=passband, borders=borders, log=log, base=base)
    integral = numpy.ma.MaskedArray(integral, mask=numpy.invert(numpy.absolute(integral) > 0))

    if yerr is None and zerr is None:
        return numpy.ma.divide(weighted_integral,integral), None

    # Construct error
    weighted_integral_err = numpy.zeros_like(weighted_integral, dtype=float)

    # Get the integrand for the y errors
    if yerr is not None:
        if yerr.shape != y.shape:
            raise ValueError('Incorrect shape for y errors.')
        ye_integrand = (z[:,None]-numpy.atleast_1d(integral)[None,:])*yerr[:,None]
        weighted_integral_err += numpy.square(passband_integral(x, ye_integrand, passband=passband,
                                                                borders=borders, log=log,
                                                                base=base, quad=True))

    # Get the integrand for the z errors
    if zerr is not None:
        if zerr.shape != z.shape:
            raise ValueError('Incorrect shape for z errors.')
        ze_integrand = _zerr*y
        weighted_integral_err += numpy.square(passband_integral(x, ze_integrand, passband=passband,
                                                                borders=borders, log=log,
                                                                base=base, quad=True))
    
    # Finalize the error
    weighted_integral_err = numpy.sqrt(weighted_integral_err)

    # Return the passband weighted mean and its error
    return numpy.ma.divide(weighted_integral, integral), \
                numpy.ma.divide(weighted_integral_err, numpy.absolute(integral))


def passband_weighted_sdev(x, y, z, passband=None, borders=False, yerr=None, zerr=None, log=False,
                           base=10.0):
    r"""
    Determine the y-weighted standard deviation of z over the passband;
    i.e.,

    .. math::

        \sigma^2 = \frac{\int_{x_1}^{x_2} y z^2 dx}{\int_{x_1}_{x_2} y
        dx} - \mu^2,
        
    where :math:`\mu` is the y-weighted mean of z over the passband (see
    :func:`passband_weighted_mean`) and the integrals are determined by
    the discrete sum; i.e., :math:`\sigma^2 = S(y z^2)/S(y) - \mu^2`
    where :math:`S(y)` is provided by :func:`passband_integral`.

    Because the error in the various components of the :math:`\sigma`
    calculation are correlated, they are calculated here by explicit
    error propagation:

    .. math::

        \epsilon_{\sigma^2}^2 = \frac{1}{S(y)^2} (S([(z-\mu)^2 + \mu^2 -
        \mu_2]^2 \epsilon_y^2) + S(4 (z-\mu)^2 y^2\epsilon_z^2)),
        
    where :math:`\mu_2` is the first term in the :math:`\sigma^2`
    calculation above and
    
    .. math::

        \epsilon_\sigma = \frac{\epsilon_{\sigma^2}}{2 \sigma}.

    Returns:
        `numpy.ma.MaskedArray`_: Two masked arrays with the
        passband-weighted standard deviation and its error. The error
        is returned as `None` if no errors are provided.
    """
    mu = passband_weighted_mean(x, y, z, passband=passband, borders=borders, log=log,
                                base=base)[0]
    mu2 = passband_weighted_mean(x, y, z*z, passband=passband, borders=borders, log=log,
                                 base=base)[0]
    sigma = numpy.ma.sqrt(mu2 - numpy.square(mu))

    if yerr is None and zerr is None:
        return sigma, None

    # Construct error
    sigma_error = numpy.zeros_like(sigma, dtype=float) if isinstance(sigma, numpy.ndarray) else 0.

    # Get the integrand for the y errors
    if yerr is not None:
        if yerr.shape != y.shape:
            raise ValueError('Incorrect shape for y errors.')
        ye_integrand = (numpy.square(z[:,None]-numpy.atleast_1d(mu)[None,:])
                            + numpy.atleast_1d(numpy.square(mu)-mu2)[None,:])*yerr[:,None]
        sigma_error += numpy.square(passband_integral(x, ye_integrand, passband=passband,
                                                      borders=borders, log=log, base=base,
                                                      quad=True))

    # Get the integrand for the z errors
    if zerr is not None:
        if zerr.shape != z.shape:
            raise ValueError('Incorrect shape for z errors.')
        ze_integrand = 2*(z[:,None]-numpy.atleast_1d(mu)[None,:])*(y*zerr)[:,None]
        sigma_error += numpy.square(passband_integral(x, ze_integrand, passband=passband,
                                                      borders=borders, log=log, base=base,
                                                      quad=True))

    # Finalize the error in sigma
    sigma_error = numpy.sqrt(sigma_error) \
                        / numpy.absolute(passband_integral(x, y, passband=passband,
                                         borders=borders, log=log, base=base) * 2*sigma)

    return sigma, sigma_error


def pseudocontinuum(x, y, passband=None, err=None, log=True, weighted_center=False):
    """
    Get the pseudocontinua in a set of passbands for a single vector (y)

    Returns:
        :obj:`float`, `numpy.ndarray`_: Return five arrays or floats:
        (1) The center of each passband, (2) the mean continuum
        level, (3) the propagated error in the continuum level (will
        be None if no errors are provided), (4) flag that part of the
        passband was masked, (5) flag that the passband was fully
        masked or empty.
    """
    # Calculate the pseudo-continua in the sidebands
    continuum, continuum_error = passband_integrated_mean(x, y, err=err, passband=passband,
                                                          log=log)
    if weighted_center:
        band_center = passband_weighted_mean(x, y, x, passband=passband, log=log)[0]
    else:
        band_center = numpy.mean(x) if passband is None else numpy.mean(passband, axis=1)

    # Calculate the fraction of the band that is covered by unmasked
    # pixels
    interval_frac = passband_integrated_width(x, y, passband=passband, log=log) \
                            / numpy.diff(passband, axis=1).ravel()

    return band_center, continuum, continuum_error, interval_frac < 1.0, \
                numpy.invert(interval_frac > 0.0)


def emission_line_equivalent_width(wave, flux, bluebands, redbands, line_centroid, line_flux,
                                   ivar=None, mask=None, log=True, redshift=None,
                                   line_flux_err=None, include_band=None):
    """
    Compute the equivalent width for emission lines provided the
    spectra and the previously measured line flux.

    Args:
        wave (`numpy.ndarray`_):
            Vector with the observed wavelength of each spectrum in
            angstroms.
        flux (`numpy.ndarray`_):
            Array (1 or 2) with the observed flux density (units in per
            angstrom) with size Nspec x Nwave.
        blueside (`numpy.ndarray`_):
            Wavelength limits for the blue sidebands in angstroms, with
            size Nbands x 2.
        redside (`numpy.ndarray`_):
            Wavelength limits for the red sidebands in angstroms, with
            size Nbands x 2.
        line_centroid (`numpy.ndarray`_):
            Wavelengths at which to sample the continuum for the
            equivalent width measurement.  Can be anything, but should
            typically be the *observed* wavelength of line center in
            angstroms, with size Nspec x Nband.
        line_flux (`numpy.ndarray`_):
            Integrated flux of the emission feature with size Nspec x
            Nband.
        ivar (`numpy.ndarray`_, optional): 
            Inverse variance in the observed flux, with shape that
            matches flux.  Default ignores error calculation.
            **Currently the equivalent width errors do not account for
            errors in the continuum characterization beneath the line.
            So this ivar array is ignored!**
        mask (`numpy.ndarray`_, optional): 
            Boolean bad-pixel mask: True values are considered to be bad
            pixels.  Default assumes all pixels are good.
        log (:obj:`bool`, optional): 
            Boolean that the spectra are logarithmically sampled in
            wavelength.
        redshift (`numpy.ndarray`_, optional):
            Redshift of each spectrum used to appropriately shift the
            definition of the bandpasses.  If a single vector is
            provided, the length must match the number of provided
            spectra; if an array is provided, the shape must be
            (Nspec,Nband).  If None, all measurements are done assuming
            the redshift is 0 (i.e., that the observed and rest frames
            are identical).
        line_flux_err (`numpy.ndarray`_, optional):
            Errors in the line flux, with size Nspec x Nband.  Default
            is to ignore the error propagation.
        include_band (`numpy.ndarray`_, optional): 
            Boolean array with size Nspec x Nband used to select which
            bands to use in the calculation for each spectrum.  Default
            is to include all bands for all spectra. 

    Returns:
        :obj:`tuple`: Returns the following arrays, all with shape
        (Nspec, Nbands):
            
            #. the passband median in the blue and red sidebands
            #. a boolean array if the sideband medians were measured
               and have positive values
            #. The continuum value used to calculate the equivalent
               width
            #. The equivalent with measurements and their errors; the
               errors are 0 if no line flux errors were provided

    Raises:
        ValueError: Raised if the shapes of the input arrays are not
            correct.

    """
    # Errors in the flux are currently not
    if ivar is not None:
        warnings.warn('Equivalent width does not propagate error from continuum measurement.')
    noise = None

    # Convert the flux to a masked array, if necessary
    if mask is not None and mask.shape != flux.shape:
        raise ValueError('Input mask must have the same shape as the flux array.')
    _flux = numpy.ma.atleast_2d(flux if isinstance(flux, numpy.ma.MaskedArray) \
                                        else numpy.ma.MaskedArray(flux, mask=mask))
    if len(wave) != _flux.shape[1]:
        raise ValueError('Wavelength vector does not match shape of the flux array.')

#    # Convert the ivar to 1-sigma error, if available
#    if ivar is None:
#        noise = None
#    else:
#        if ivar.shape != _flux.shape:
#            raise ValueError('Input ivar array must be the same shape as the flux array.')
#        _ivar = ivar if isinstance(ivar, numpy.ma.MaskedArray) else \
#                    numpy.ma.MaskedArray(ivar, mask=numpy.ma.getmaskarray(_flux))
#        noise = numpy.ma.sqrt(1.0 /_ivar)

    # Check band definitions
    if bluebands.shape[1] != 2:
        raise ValueError('Input bands must have shape Nband x 2.')
    if bluebands.shape != redbands.shape:
        raise ValueError('Input side bands do not have the same shape.')

    # Check the shapes
    nspec = flux.shape[0]
    nbands = bluebands.shape[0]

    expected_shape = (nspec, nbands)
    if line_centroid.shape != expected_shape:
        raise ValueError('Line centroid array must have shape: {0}'.format(expected_shape))
    if line_flux.shape != expected_shape:
        raise ValueError('Line flux array must have shape: {0}'.format(expected_shape))
    if line_flux_err is not None and line_flux_err.shape != expected_shape:
        raise ValueError('Line flux error array must have shape: {0}'.format(expected_shape))

    # Check the input band flags
    _include_band = numpy.ones(expected_shape, dtype=bool) \
                        if include_band is None else include_band
    if _include_band.shape != expected_shape:
        raise ValueError('Bands flags array must have shape: {0}'.format(expected_shape))

    # Check the input redshifts
    _redshift = numpy.zeros(expected_shape, dtype=float) if redshift is None else redshift
    if _redshift.ndim == 1:
        if len(_redshift) != nspec:
            raise ValueError('Must provide at least one redshift per input spectrum.')
        _redshift = numpy.array([_redshift]*nbands).T
    if _redshift.ndim == 2 and _redshift.shape != expected_shape:
        raise ValueError('Provided redshift array does not match the number of spectra and bands.')

    # Initialize the output data
    bmed = numpy.zeros(expected_shape, dtype=float)
    rmed = numpy.zeros(expected_shape, dtype=float)
    pos = _include_band.copy()
    ewcont = numpy.zeros(expected_shape, dtype=float)
    ew = numpy.zeros(expected_shape, dtype=float)
    ewerr = numpy.zeros(expected_shape, dtype=float)

    # Measure the pseudo-continuum in the sidebands
    for i in range(nspec):

        print('Measuring emission-line equivalent widths in spectrum: {0}/{1}'.format(i+1,nspec),
              end='\r')

        # Spectra to use
        spec = _flux[i,:]
        _noise = None if noise is None else noise[i,:]

        # Bands to use
        indx = _include_band[i,:]

        # Shift the bands to the appropriate redshift
        _bluebands = bluebands[indx,:]*(1.0+_redshift[i,indx,None])
        _redbands = redbands[indx,:]*(1.0+_redshift[i,indx,None])

        # Center of each band (for all bands!)
        bcen = numpy.mean(bluebands*(1.0+_redshift[i,:,None]), axis=1)
        rcen = numpy.mean(redbands*(1.0+_redshift[i,:,None]), axis=1)

        # Median of each band
        bmed[i,indx] = passband_median(wave, _flux[i,:], passband=_bluebands)
        rmed[i,indx] = passband_median(wave, _flux[i,:], passband=_redbands)

        # Construct the continuum level at the center of the main
        # band
        pos[i,:] = indx & (bmed[i,:] > 0) & (rmed[i,:] > 0)

        # Continuum at line centroid
        m = (rmed[i,pos[i,:]] - bmed[i,pos[i,:]])/(rcen[pos[i,:]] - bcen[pos[i,:]])
        ewcont[i,pos[i,:]] = m*(line_centroid[i,pos[i,:]] - bcen[pos[i,:]]) + bmed[i,pos[i,:]]

        # Compute the equivalent width
        ew[i,pos[i,:]] = line_flux[i,pos[i,:]] / ewcont[i,pos[i,:]]
        ewerr[i,pos[i,:]] = 0.0 if line_flux_err is None \
                                else line_flux_err[i,pos[i,:]] / ewcont[i,pos[i,:]]

        # Correct for the redshift
        ew[i,pos[i,:]] /= (1.0+_redshift[i,pos[i,:]])
        ewerr[i,pos[i,:]] /= (1.0+_redshift[i,pos[i,:]])

#        pyplot.step(wave, _flux[i,:], where='mid', color='k', lw=0.5, zorder=1)
#        pyplot.scatter(bcen[indx], bmed[i,indx], marker='.', s=50, color='b', lw=0, zorder=3)
#        pyplot.scatter(rcen[indx], rmed[i,indx], marker='.', s=50, color='r', lw=0, zorder=3)
#        pyplot.scatter(line_centroid[i,indx], ewcont[i,indx], marker='.', s=50, color='g',
#                       lw=0, zorder=3)
#        pyplot.show()

    print('Measuring emission-line equivalent widths in spectrum: {0}/{0}'.format(nspec))
    return bmed, rmed, pos, ewcont, ew, ewerr





#!/usr/bin/env python3

import time
import os
import astropy.constants
from astropy.io import fits

from mangadap.config import manga
from mangadap.survey.manga_dap import manga_dap
from mangadap.par.obsinput import ObsInputPar
from mangadap.par.analysisplan import AnalysisPlan, AnalysisPlanSet

#-----------------------------------------------------------------------------
def get_obsinput(plt, ifu, drpall_file=None):
    """
    Grab the input parameters the DAP requires for each observation to
    fit a cube.  If the drpall file is None, use the default path.
    """
    if drpall_file is None:
        drpall_file = manga.drpall_file()
    hdu = fits.open(drpall_file)
    indx = hdu[1].data['PLATEIFU'] == '{0}-{1}'.format(plt, ifu)

    return ObsInputPar(plate=plt, ifudesign=ifu, mode='CUBE',
                       vel=astropy.constants.c.to('km/s').value*hdu[1].data['NSA_Z'][indx][0],
                       vdisp=100.,
                       ell=1-hdu[1].data['NSA_ELPETRO_BA'][indx][0],
                       pa=hdu[1].data['NSA_ELPETRO_PHI'][indx][0],
                       reff=hdu[1].data['NSA_ELPETRO_TH50_R'][indx][0])


def fit_one_cube(plt, ifu, drpall_file=None, directory_path=None, analysis_path=None):
    # Grab the required input parameters
    obs = get_obsinput(plt, ifu, drpall_file='./data/drpall-v2_4_3.fits')

    # Define how you want to analyze the data
    plan = AnalysisPlanSet([ AnalysisPlan(drpqa_key='SNRG',
                                          bin_key='VOR10',
                                          continuum_key='GAU-MILESHC',
                                          elmom_key='EMOMM',
                                          elfit_key='EFITM-MIUSCATTHIN',
                                          spindex_key='INDXEN') ])
    # Run it!
    return manga_dap(obs, plan, verbose=2, directory_path='./data', analysis_path='./output')


#-----------------------------------------------------------------------------

if __name__ == '__main__':
    t = time.perf_counter()

    fit_one_cube(7815, 3702, drpall_file='./data/drpall-v2_4_3.fits', directory_path='./data',
                 analysis_path='./output')

    print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))




#!/usr/bin/env python3

import os
import time
import argparse

from IPython import embed

import numpy
from astropy.io import fits
import astropy.constants
from matplotlib import pyplot, ticker

from mangadap.util.constants import DAPConstants
from mangadap.util.filter import interpolate_masked_vector
from mangadap.util.drpfits import DRPFitsBitMask
from mangadap.util.fileio import compress_file
from mangadap.util.resolution import SpectralResolution
from mangadap.util.pixelmask import SpectralPixelMask

from mangadap.par.artifactdb import ArtifactDB
from mangadap.par.emissionmomentsdb import EmissionMomentsDB
from mangadap.par.emissionlinedb import EmissionLineDB

from mangadap.proc.bandpassfilter import emission_line_equivalent_width

from mangadap.proc.templatelibrary import TemplateLibrary
from mangadap.proc.emissionlinemoments import EmissionLineMoments, EmissionLineMomentsBitMask
from mangadap.proc.sasuke import Sasuke
from mangadap.proc.ppxffit import PPXFFit
from mangadap.proc.stellarcontinuummodel import StellarContinuumModel, StellarContinuumModelBitMask
from mangadap.proc.emissionlinemodel import EmissionLineModelBitMask
from mangadap.proc.spectralfitting import EmissionLineFit

#-----------------------------------------------------------------------------
def object_data(f, fit_flag_file):

    with fits.open(f) as hdu:
        flux = hdu['FLUX'].data
        ivar = hdu['IVAR'].data
        mask = hdu['MASK'].data
        sres = hdu['LSF'].data
        wave = hdu['WAVE'].data
        redshift = hdu['Z'].data

    sres = numpy.ma.divide(wave[None,:], sres) / DAPConstants.sig2fwhm
    if numpy.any(numpy.ma.getmaskarray(sres)):
        sres = numpy.ma.apply_along_axis(interpolate_masked_vector, 1, sres)

    bm = DRPFitsBitMask()
    nspec = flux.shape[0]

    fit_flags = numpy.ones(nspec, dtype=bool) if fit_flag_file is None \
                    else (numpy.genfromtxt(fit_flag_file)[:,3]).astype(bool)
    if len(fit_flags) != nspec:
        print(nspec, len(fit_flags))
        raise ValueError('Incorrect number of fitting flags.')

    ferr = numpy.ma.power(ivar, -0.5)
    bool_mask = bm.flagged(mask, flag=['DONOTUSE', 'FORESTAR']) | numpy.ma.getmaskarray(ferr)
    
    return wave, numpy.ma.MaskedArray(flux, mask=bool_mask), \
                numpy.ma.MaskedArray(ferr, mask=bool_mask), sres, redshift, fit_flags

def init_ax(fig, pos):
    ax = fig.add_axes(pos) #, facecolor='0.95')
    ax.minorticks_on()
    ax.tick_params(which='major', length=10)
    ax.tick_params(which='minor', length=5)
    ax.tick_params(which='both', direction='out', top='on', right='on')
    return ax

def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument('inp', type=str, help='input fits file with spectra to fit')
    parser.add_argument('out', type=str, help='output fits file with fitted models')

    parser.add_argument('--output_root', type=str, help='root directory for output', default=None)
    parser.add_argument('--sc_tpl', type=str, default='MILESHC',
                        help='template library to use for the stellar kinematics')
    parser.add_argument('--sc_vsr', type=int, default=4,
                        help='velocity-scale ratio for templates during stellar kinematics fit')
    parser.add_argument('--sc_deg', type=int, default=8,
                        help='additive order during stellar kinematics fit')
    parser.add_argument('--sc_only', default=False, action='store_true',
                        help='skip the emission-line modeling and only fit the stellar kinematics')

    parser.add_argument('--el_tpl', type=str, default='MILESHC',
                        help='template library to use for the emission-line fitting')
    parser.add_argument('--el_band', type=str, default='ELBMPL9',
                        help='keyword with the emission-line bandpasses for the moment analysis.')
    parser.add_argument('--el_list', type=str, default='ELPMPL9',
                        help='keyword with the emission-line database to fit')
    parser.add_argument('--el_vsr', type=int, default=4,
                        help='velocity-scale ratio for templates during emission-line fit')
    parser.add_argument('--el_deg', type=int, default=8,
                        help='multiplicative order during emission-line fit')

    parser.add_argument('--spec_flags', type=str, default=None,
                        help='fourth column sets whether (1) or not (0) to fit spectrum')

    return parser.parse_args()
    

def write(ofile, wave, cont_flux, cont_mask, cont_par, model_flux=None, model_mask=None,
          eml_flux=None, eml_fit_par=None, eml_eml_par=None, elmom=None):

    compress = False
    if ofile.split('.')[-1] == 'gz':
        _ofile = ofile[:ofile.rfind('.')] 
        compress = True
    else:
        _ofile = ofile

    hdu = fits.HDUList([fits.PrimaryHDU(),
                        fits.ImageHDU(data=wave, name='WAVE'),
                        fits.ImageHDU(data=cont_flux.data, name='STELLAR'),
                        fits.ImageHDU(data=cont_mask, name='STELLAR_MASK'),
                        cont_par.to_hdu(name='STRPAR')])

    if model_flux is not None:
        hdu += [fits.ImageHDU(data=model_flux.data, name='MODEL'),
                fits.ImageHDU(data=model_mask, name='MODEL_MASK'),
                fits.ImageHDU(data=eml_flux.data, name='EMLINE'),
                eml_fit_par.to_hdu(name='EMLFIT'),
                eml_eml_par.to_hdu(name='EMLPAR'),
                elmom.to_hdu(name='EMLMOM')]

    hdu.writeto(_ofile, overwrite=True)
    
    # Compress if desired
    if compress:
        compress_file(_ofile, clobber=True)
        os.remove(_ofile)


def main():
    t = time.perf_counter()
    arg = parse_args()
    if not os.path.isfile(arg.inp):
        raise FileNotFoundError('No file: {0}'.format(arg.inp))
    directory_path = os.getcwd() if arg.output_root is None else os.path.abspath(arg.output_root)
    if not os.path.isdir(directory_path):
        os.makedirs(directory_path)

    data_file = os.path.abspath(arg.inp)
    fit_file = os.path.join(directory_path, arg.out)
    flag_db = None if arg.spec_flags is None else os.path.abspath(arg.spec_flags)

    # Read the data
    spectral_step = 1e-4
    wave, flux, ferr, sres, redshift, fit_spectrum = object_data(data_file, flag_db)
    nspec, npix = flux.shape
    dispersion = numpy.full(nspec, 100., dtype=numpy.float)

#    fit_spectrum[:] = False
#    fit_spectrum[0] = True
#    fit_spectrum[171] = True
#    fit_spectrum[791] = True

    # Mask spectra that should not be fit
    indx = numpy.any(numpy.logical_not(numpy.ma.getmaskarray(flux)), axis=1) & fit_spectrum
    flux[numpy.logical_not(indx),:] = numpy.ma.masked

    print('Read: {0}'.format(arg.inp))
    print('Contains {0} spectra'.format(nspec))
    print(' each with {0} pixels'.format(npix))
    print('Fitting {0} spectra.'.format(numpy.sum(fit_spectrum)))

    #-------------------------------------------------------------------
    #-------------------------------------------------------------------
    # Fit the stellar continuum

    # Construct the template library
    sc_tpl = TemplateLibrary(arg.sc_tpl, match_resolution=False, velscale_ratio=arg.sc_vsr,
                             spectral_step=spectral_step, log=True, hardcopy=False)
    # Set the spectral resolution
    sc_tpl_sres = numpy.mean(sc_tpl['SPECRES'].data, axis=0).ravel()
    # Set the pixel mask
    sc_pixel_mask = SpectralPixelMask(artdb=ArtifactDB.from_key('BADSKY'),
                                      emldb=EmissionLineDB.from_key('ELPMPL8'))
    # Instantiate the fitting class
    ppxf = PPXFFit(StellarContinuumModelBitMask())

    # The following call performs the fit to the spectrum. Specifically
    # note that the code only fits the first two moments, uses an
    # 8th-order additive polynomial, and uses the 'no_global_wrej'
    # iteration mode. See
    # https://sdss-mangadap.readthedocs.io/en/latest/api/mangadap.proc.ppxffit.html#mangadap.proc.ppxffit.PPXFFit.fit
    cont_wave, cont_flux, cont_mask, cont_par \
        = ppxf.fit(sc_tpl['WAVE'].data.copy(), sc_tpl['FLUX'].data.copy(), wave, flux, ferr,
                   redshift, dispersion, iteration_mode='no_global_wrej', reject_boxcar=100,
                   ensemble=False, velscale_ratio=arg.sc_vsr, mask=sc_pixel_mask,
                   matched_resolution=False, tpl_sres=sc_tpl_sres, obj_sres=sres,
                   degree=arg.sc_deg, moments=2) #, plot=True)

    if arg.sc_only:
        write(fit_file, wave, cont_flux, cont_mask, cont_par)
        print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))
        return

#    if numpy.any(cont_par['KIN'][:,1] < 0):
#        embed()
#        exit()
    #-------------------------------------------------------------------


    #-------------------------------------------------------------------
    #-------------------------------------------------------------------
    # Measure the emission-line moments

#    # Remask the continuum fit
#    sc_continuum = StellarContinuumModel.reset_continuum_mask_window(
#                        numpy.ma.MaskedArray(cont_flux, mask=cont_mask>0))
#    # Read the database that define the emission lines and passbands
#    momdb = EmissionMomentsDB.from_key(arg.el_band)
#    # Measure the moments
#    elmom = EmissionLineMoments.measure_moments(momdb, wave, flux, continuum=sc_continuum,
#                                                redshift=redshift)
    #-------------------------------------------------------------------

    #-------------------------------------------------------------------
    # Fit the emission-line model

    # Set the emission-line continuum templates if different from those
    # used for the stellar continuum
    if arg.sc_tpl == arg.el_tpl:
        # If the keywords are the same, just copy over the previous
        # library and the best fitting stellar kinematics
        el_tpl = sc_tpl
        el_tpl_sres = sc_tpl_sres
        stellar_kinematics = cont_par['KIN'].copy()
    else:
        # If the template sets are different, we need to match the
        # spectral resolution to the galaxy data and use the corrected
        # velocity dispersions.
        _sres = SpectralResolution(wave, sres[0,:], log10=True)
        el_tpl = TemplateLibrary(arg.el_tpl, sres=_sres, velscale_ratio=arg.el_vsr,
                                 spectral_step=spectral_step, log=True, hardcopy=False)
        el_tpl_sres = numpy.mean(el_tpl['SPECRES'].data, axis=0).ravel()
        stellar_kinematics = cont_par['KIN'].copy()
        stellar_kinematics[:,1] = numpy.ma.sqrt(numpy.square(cont_par['KIN'][:,1]) -
                                        numpy.square(cont_par['SIGMACORR_SRES'])).filled(0.0)

#    if numpy.any(cont_par['KIN'][:,1] < 0):
#        embed()
#        exit()
#
#    if numpy.any(stellar_kinematics[:,1] < 0):
#        embed()
#        exit()

    # Mask the 5577 sky line
    el_pixel_mask = SpectralPixelMask(artdb=ArtifactDB.from_key('BADSKY'))

    # Read the emission line fitting database
    emldb = EmissionLineDB.from_key(arg.el_list)

    # Instantiate the fitting class
    emlfit = Sasuke(EmissionLineModelBitMask())

    # TODO: Improve the initial velocity guess using the first moment...

    # Perform the fit
    elfit_time = time.perf_counter()
    model_wave, model_flux, eml_flux, model_mask, eml_fit_par, eml_eml_par \
            = emlfit.fit(emldb, wave, flux, obj_ferr=ferr, obj_mask=el_pixel_mask, obj_sres=sres,
                         guess_redshift=redshift, guess_dispersion=dispersion, reject_boxcar=101,
                         stpl_wave=el_tpl['WAVE'].data, stpl_flux=el_tpl['FLUX'].data,
                         stpl_sres=el_tpl_sres, stellar_kinematics=stellar_kinematics,
                         etpl_sinst_mode='offset', etpl_sinst_min=10., velscale_ratio=arg.el_vsr,
                         matched_resolution=False, mdegree=arg.el_deg, ensemble=False)#, plot=True)
    print('EML FIT TIME: ', time.perf_counter() - elfit_time)

    # Line-fit metrics (should this be done in the fit method?)
    eml_eml_par = EmissionLineFit.line_metrics(emldb, wave, flux, ferr, model_flux, eml_eml_par,
                                               model_mask=model_mask, bitmask=emlfit.bitmask)

    # Equivalent widths
    EmissionLineFit.measure_equivalent_width(wave, flux, emldb, eml_eml_par,
                                             bitmask=emlfit.bitmask, checkdb=False)

    # Measure the emission-line moments
    #   - Model continuum
    continuum = StellarContinuumModel.reset_continuum_mask_window(model_flux - eml_flux)
    #   - Updated redshifts
    fit_redshift = eml_eml_par['KIN'][:,numpy.where(emldb['name'] == 'Ha')[0][0],0] \
                        / astropy.constants.c.to('km/s').value
    #   - Set the moment database
    momdb = EmissionMomentsDB.from_key(arg.el_band)
    #   - Set the moment bitmask
    mombm = EmissionLineMomentsBitMask()
    #   - Measure the moments
    elmom = EmissionLineMoments.measure_moments(momdb, wave, flux, ivar=numpy.ma.power(ferr, -2),
                                                continuum=continuum, redshift=fit_redshift,
                                                bitmask=mombm)
    #   - Select the bands that are valid
    include_band = numpy.array([numpy.logical_not(momdb.dummy)]*nspec) \
                        & numpy.logical_not(mombm.flagged(elmom['MASK'],
                                                          flag=['BLUE_EMPTY', 'RED_EMPTY']))
    #   - Set the line center at the center of the primary passband
    line_center = (1.0+fit_redshift)[:,None]*momdb['restwave'][None,:]
    elmom['BMED'], elmom['RMED'], pos, elmom['EWCONT'], elmom['EW'], elmom['EWERR'] \
            = emission_line_equivalent_width(wave, flux, momdb['blueside'], momdb['redside'],
                                             line_center, elmom['FLUX'], redshift=fit_redshift,
                                             line_flux_err=elmom['FLUXERR'],
                                             include_band=include_band)
    #   - Flag non-positive measurements
    indx = include_band & numpy.logical_not(pos)
    elmom['MASK'][indx] = mombm.turn_on(elmom['MASK'][indx], 'NON_POSITIVE_CONTINUUM')
    #   - Set the binids
    elmom['BINID'] = numpy.arange(nspec)
    elmom['BINID_INDEX'] = numpy.arange(nspec)

    write(fit_file, wave, cont_flux, cont_mask, cont_par, model_flux=model_flux,
          model_mask=model_mask, eml_flux=eml_flux, eml_fit_par=eml_fit_par,
          eml_eml_par=eml_eml_par, elmom=elmom)
    print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))


if __name__ == '__main__':
    main()